# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for tasks."""
import os
import shutil
import tempfile
import time
import unittest

import mock

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import pubsub
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class RedoTestcaseTest(unittest.TestCase):
  """Test redo_testcase()."""

  def test_invalid_task(self):
    """Raise an exception on an invalid task."""
    with self.assertRaises(tasks.InvalidRedoTask) as cm:
      tasks.redo_testcase(None, ['blame', 'rand'], 'test@user.com')

    self.assertEqual("The task 'rand' is invalid.", str(cm.exception))


@test_utils.integration
@test_utils.with_cloud_emulators('datastore', 'pubsub')
class GetTaskTest(unittest.TestCase):
  """GetTask tests."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.base.persistent_cache.get_value',
        'clusterfuzz._internal.base.persistent_cache.set_value',
        'clusterfuzz._internal.base.utils.utcnow',
        'time.sleep',
    ])

    self.mock.get_value.return_value = None
    self.mock.sleep.return_value = None
    data_types.Job(name='job').put()

    client = pubsub.PubSubClient()
    topic = pubsub.topic_name('test-clusterfuzz', 'jobs-linux')
    client.create_topic(topic)
    client.create_subscription(
        pubsub.subscription_name('test-clusterfuzz', 'jobs-linux'), topic)

    topic = pubsub.topic_name('test-clusterfuzz', 'high-end-jobs-linux')
    client.create_topic(topic)
    client.create_subscription(
        pubsub.subscription_name('test-clusterfuzz', 'high-end-jobs-linux'),
        topic)

    self.mock.utcnow.return_value = test_utils.CURRENT_TIME.replace(
        microsecond=0)

  def test_high_end(self):
    """Test high end tasks."""
    environment.set_value('THREAD_MULTIPLIER', 4)
    tasks.add_task(
        'test', 'high', 'job', queue='high-end-jobs-linux', wait_time=0)
    tasks.add_task('test', 'normal', 'job', queue='jobs-linux', wait_time=0)

    task = tasks.get_task()
    self.assertEqual('test', task.command)
    self.assertEqual('high', task.argument)
    self.assertEqual('job', task.job)
    self.assertEqual('test high job', task.payload())

  def test_regular(self):
    """Test regular tasks."""
    environment.set_value('THREAD_MULTIPLIER', 1)
    tasks.add_task(
        'test', 'high', 'job', queue='high-end-jobs-linux', wait_time=0)
    tasks.add_task('test', 'normal', 'job', queue='jobs-linux', wait_time=0)

    task = tasks.get_task()
    self.assertEqual('test', task.command)
    self.assertEqual('normal', task.argument)
    self.assertEqual('job', task.job)
    self.assertEqual('test normal job', task.payload())

  def test_preemptible(self):
    """Test preemptible bot tasks."""
    environment.set_value('PREEMPTIBLE', True)
    environment.set_value('THREAD_MULTIPLIER', 1)
    tasks.add_task(
        'test', 'high', 'job', queue='high-end-jobs-linux', wait_time=0)
    tasks.add_task('test', 'normal', 'job', queue='jobs-linux', wait_time=0)

    task = tasks.get_task()
    self.assertIsNone(task)

  def test_defer(self):
    """Test deferring tasks which shouldn't be run yet."""
    tasks.add_task('test', 'normal1', 'job', wait_time=60)
    tasks.add_task('test', 'normal2', 'job', wait_time=600)
    tasks.add_task('test', 'normal3', 'job', wait_time=700)
    tasks.add_task('test', 'normal4', 'job', wait_time=0)

    with mock.patch.object(pubsub.ReceivedMessage,
                           'modify_ack_deadline') as mock_modify:
      task = tasks.get_task()
      self.assertEqual('test', task.command)
      self.assertEqual('normal4', task.argument)
      self.assertEqual('job', task.job)
      self.assertEqual('test normal4 job', task.payload())

      self.assertEqual(3, mock_modify.call_count)
      mock_modify.assert_has_calls([
          mock.call(60),
          mock.call(600),
          mock.call(600),
      ])

  def test_command_override(self):
    """Test command override."""
    environment.set_value('COMMAND_OVERRIDE', 'test override job')
    tasks.add_task('test', 'normal', 'job', wait_time=0)

    task = tasks.get_task()
    self.assertEqual('test', task.command)
    self.assertEqual('override', task.argument)
    self.assertEqual('job', task.job)
    self.assertEqual('test override job', task.payload())


class LeaseTaskTest(unittest.TestCase):
  """Tests for leasing tasks."""

  def setUp(self):
    helpers.patch_environ(self)

    helpers.patch(self, [
        'clusterfuzz._internal.datastore.data_handler.update_heartbeat',
        'time.time',
    ])

    self.temp_dir = tempfile.mkdtemp()
    os.environ['CACHE_DIR'] = self.temp_dir
    self.mock.time.return_value = 1337

  def tearDown(self):
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  def test_lease_finish_before_latest(self):
    """Test leasing a task and finishing before the latest time."""
    message = mock.MagicMock()
    message.attributes = {
        'command': 'cmd',
        'argument': 'arg',
        'job': 'job',
        'eta': time.time(),
    }

    task = tasks.PubSubTask(message)
    self.mock.time.side_effect = [0, 0, 0, 600, 1200]

    with task.lease() as thread:
      self.assertEqual('21600', os.environ['TASK_LEASE_SECONDS'])

    self.assertEqual(1, message.modify_ack_deadline.call_count)
    message.modify_ack_deadline.assert_has_calls([
        mock.call(600),
    ])

    self.assertFalse(thread.is_alive())
    self.assertEqual(1, message.ack.call_count)

  def test_lease_reach_latest(self):
    """Test leasing a task and reaching the maximum lease time."""
    message = mock.MagicMock()
    message.attributes = {
        'command': 'cmd',
        'argument': 'arg',
        'job': 'job',
        'eta': time.time(),
    }

    task = tasks.PubSubTask(message)

    event = mock.MagicMock()
    event.wait.side_effect = [False, False, False]
    self.mock.time.side_effect = [
        0, 0, 0, 600, tasks.TASK_LEASE_SECONDS - 30, tasks.TASK_LEASE_SECONDS
    ]

    with task.lease(_event=event) as thread:
      self.assertEqual('21600', os.environ['TASK_LEASE_SECONDS'])
      thread.join()

    self.assertEqual(3, message.modify_ack_deadline.call_count)
    # Last extension should be until TASK_LEASE_SECONDS.
    message.modify_ack_deadline.assert_has_calls([
        mock.call(600),
        mock.call(600),
        mock.call(30),
    ])

    self.assertEqual(1, message.ack.call_count)
    self.assertFalse(thread.is_alive())

  def test_lease_exception(self):
    """Test lease with an exception during the task."""
    message = mock.MagicMock()
    message.attributes = {
        'command': 'cmd',
        'argument': 'arg',
        'job': 'job',
        'eta': time.time(),
    }

    task = tasks.PubSubTask(message)

    class Error(Exception):
      """Fake error."""

    with self.assertRaises(Error):
      with task.lease() as thread:
        raise Error

    self.assertFalse(thread.is_alive())
    self.assertEqual(0, message.ack.call_count)


@test_utils.with_cloud_emulators('datastore')
class QueueForTestcaseTest(unittest.TestCase):
  """Tests for queue_for_testcase."""

  def setUp(self):
    helpers.patch_environ(self)

    data_types.Job(name='job_linux', platform='LINUX').put()
    data_types.Job(name='job_project', platform='PROJECT_LINUX_LIB').put()

  def test_regular_queue(self):
    """Test testcase with regular queue."""
    t = data_types.Testcase(job_type='job_linux', queue='old')
    self.assertEqual('jobs-linux', tasks.queue_for_testcase(t))

    t = data_types.Testcase(job_type='job_project', queue='old')
    self.assertEqual('jobs-project-linux-lib', tasks.queue_for_testcase(t))

  def test_high_end_queue(self):
    """Test testcase with high end queue."""
    t = data_types.Testcase(job_type='job_linux', queue='high-end-jobs')
    self.assertEqual('high-end-jobs-linux', tasks.queue_for_testcase(t))

    t = data_types.Testcase(job_type='job_project', queue='high-end-jobs')
    self.assertEqual('high-end-jobs-project-linux-lib',
                     tasks.queue_for_testcase(t))


@test_utils.with_cloud_emulators('datastore', 'pubsub')
class ExternalTasksTest(unittest.TestCase):
  """Tests for adding external tasks."""

  def setUp(self):
    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.blobs.read_key',
    ])

    self.mock.read_key.return_value = b'reproducer'

    self.client = pubsub.PubSubClient()
    self.topic = pubsub.topic_name('proj', 'repro')
    self.client.create_topic(self.topic)
    self.subscription = pubsub.subscription_name('proj', 'repro')
    self.client.create_subscription(self.subscription, self.topic)

    data_types.Job(
        name='libfuzzer_asan_blah_external',
        platform='LINUX',
        environment_string=(
            'RELEASE_BUILD_BUCKET_PATH = gs://bucket/a/b/release-([0-9]+).zip\n'
            'PROJECT_NAME = proj\n'),
        external_reproduction_topic=self.topic,
        external_updates_subscription='projects/proj/subscriptions/updates'
    ).put()

    data_types.Job(
        name='libfuzzer_msan_blah_external',
        platform='LINUX',
        environment_string=('FUZZ_TARGET_BUILD_BUCKET_PATH = '
                            'gs://bucket/a/b/%TARGET%/release-([0-9]+).zip\n'
                            'PROJECT_NAME = proj\n'),
        external_reproduction_topic=self.topic,
        external_updates_subscription='projects/proj/subscriptions/updates'
    ).put()

    data_types.FuzzTarget(
        id='libFuzzer_abc', engine='libFuzzer', binary='abc').put()

    self.testcase_0 = data_types.Testcase(
        fuzzer_name='libFuzzer',
        overridden_fuzzer_name='libFuzzer_abc',
        crash_revision=1336,
        minimized_keys='key')
    self.testcase_0.set_metadata('last_tested_revision', 1337)
    self.testcase_0.put()

    self.testcase_1 = data_types.Testcase(
        fuzzer_name='libFuzzer',
        overridden_fuzzer_name='libFuzzer_abc',
        crash_revision=1336,
        fuzzed_keys='key')
    self.testcase_1.put()

  def test_progression_0(self):
    """Test adding a ASAN progression (reproduction) task."""
    tasks.add_task('progression', self.testcase_0.key.id(),
                   'libfuzzer_asan_blah_external')

    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(1, len(messages))

    message = messages[0]
    self.assertEqual(b'reproducer', message.data)
    self.assertDictEqual({
        'buildPath': 'gs://bucket/a/b/release-([0-9]+).zip',
        'fuzzer': 'libFuzzer',
        'job': 'libfuzzer_asan_blah_external',
        'minRevisionAbove': '1337',
        'project': 'proj',
        'sanitizer': 'address',
        'target': 'abc',
        'testcaseId': str(self.testcase_0.key.id()),
    }, message.attributes)

  def test_progression_1(self):
    """Test adding an MSAN progression (reproduction) task."""
    tasks.add_task('progression', self.testcase_1.key.id(),
                   'libfuzzer_msan_blah_external')

    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(1, len(messages))

    message = messages[0]
    self.assertEqual(b'reproducer', message.data)
    self.assertDictEqual({
        'buildPath': 'gs://bucket/a/b/%TARGET%/release-([0-9]+).zip',
        'fuzzer': 'libFuzzer',
        'job': 'libfuzzer_msan_blah_external',
        'minRevisionAbove': '1336',
        'project': 'proj',
        'sanitizer': 'memory',
        'target': 'abc',
        'testcaseId': str(self.testcase_1.key.id()),
    }, message.attributes)

  def test_not_progression(self):
    """Test trying to add a task that isn't progression."""
    tasks.add_task('impact', self.testcase_1.key.id(),
                   'libfuzzer_msan_blah_external')

    messages = self.client.pull_from_subscription(
        self.subscription, max_messages=1)
    self.assertEqual(0, len(messages))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Performance stats constants and helpers for libFuzzer."""

import re

from clusterfuzz._internal.bot.fuzzers import dictionary_manager
from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.bot.fuzzers.libFuzzer import constants
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz.stacktraces import constants as stacktrace_constants

# Regular expressions to detect different types of crashes.
LEAK_TESTCASE_REGEX = re.compile(r'.*ERROR: LeakSanitizer.*')
LIBFUZZER_BAD_INSTRUMENTATION_REGEX = re.compile(
    r'.*ERROR:.*Is the code instrumented for coverage.*')
LIBFUZZER_CRASH_TYPE_REGEX = r'.*Test unit written to.*{type}'
LIBFUZZER_CRASH_START_MARKER = r'.*ERROR: (libFuzzer|.*Sanitizer):'
LIBFUZZER_ANY_CRASH_TYPE_REGEX = re.compile(
    r'(%s|%s)' % (LIBFUZZER_CRASH_START_MARKER,
                  LIBFUZZER_CRASH_TYPE_REGEX.format(type='')))
LIBFUZZER_CRASH_TESTCASE_REGEX = re.compile(
    LIBFUZZER_CRASH_TYPE_REGEX.format(type='crash'))
LIBFUZZER_OOM_TESTCASE_REGEX = re.compile(
    LIBFUZZER_CRASH_TYPE_REGEX.format(type='oom'))
LIBFUZZER_SLOW_UNIT_TESTCASE_REGEX = re.compile(
    LIBFUZZER_CRASH_TYPE_REGEX.format(type='slow-unit'))
LIBFUZZER_TIMEOUT_TESTCASE_REGEX = re.compile(
    LIBFUZZER_CRASH_TYPE_REGEX.format(type='timeout'))

# Regular expressions to detect different sections of logs.
LIBFUZZER_FUZZING_STRATEGIES = re.compile(r'cf::fuzzing_strategies:\s*(.*)')
LIBFUZZER_LOG_DICTIONARY_REGEX = re.compile(r'Dictionary: \d+ entries')
LIBFUZZER_LOG_END_REGEX = re.compile(r'Done\s+\d+\s+runs.*')
LIBFUZZER_LOG_IGNORE_REGEX = re.compile(r'.*WARNING:.*Sanitizer')
LIBFUZZER_LOG_LINE_REGEX = re.compile(
    r'^#\d+[\s]*(READ|INITED|NEW|pulse|REDUCE|RELOAD|DONE|:)\s.*')
LIBFUZZER_LOG_SEED_CORPUS_INFO_REGEX = re.compile(
    r'INFO:\s+seed corpus:\s+files:\s+(\d+).*rss:\s+(\d+)Mb.*')
LIBFUZZER_LOG_START_INITED_REGEX = re.compile(
    r'(#\d+\s+INITED\s+|INFO:\s+-fork=\d+:\s+fuzzing in separate process).*')
LIBFUZZER_MERGE_LOG_STATS_REGEX = re.compile(
    r'MERGE-OUTER:\s+\d+\s+new files with'
    r'\s+(\d+)\s+new features added;'
    r'\s+(\d+)\s+new coverage edges.*')
LIBFUZZER_MODULES_LOADED_REGEX = re.compile(
    r'^INFO:\s+Loaded\s+\d+\s+(modules|PC tables)\s+\((\d+)\s+.*\).*')

# Regular expressions to extract different values from the log.
LIBFUZZER_LOG_MAX_LEN_REGEX = re.compile(
    r'.*-max_len is not provided; libFuzzer will not generate inputs larger'
    r' than (\d+) bytes.*')


def calculate_log_lines(log_lines):
  """Calculate number of logs lines of different kind in the given log."""
  # Counters to be returned.
  libfuzzer_lines_count = 0
  other_lines_count = 0
  ignored_lines_count = 0

  lines_after_last_libfuzzer_line_count = 0
  libfuzzer_inited = False
  found_libfuzzer_crash = False
  for line in log_lines:
    if not libfuzzer_inited:
      # Skip to start of libFuzzer log output.
      if LIBFUZZER_LOG_START_INITED_REGEX.match(line):
        libfuzzer_inited = True
      else:
        ignored_lines_count += 1
        continue

    if LIBFUZZER_LOG_IGNORE_REGEX.match(line):
      # We should ignore lines like sanitizer warnings, etc.
      ignored_lines_count += 1
      continue

    if LIBFUZZER_ANY_CRASH_TYPE_REGEX.match(line):
      # We should ignore whole block if a libfuzzer crash is found.
      # E.g. slow units.
      found_libfuzzer_crash = True
    elif LIBFUZZER_LOG_LINE_REGEX.match(line):
      if found_libfuzzer_crash:
        # Ignore previous block.
        other_lines_count -= lines_after_last_libfuzzer_line_count
        ignored_lines_count += lines_after_last_libfuzzer_line_count

      libfuzzer_lines_count += 1
      lines_after_last_libfuzzer_line_count = 0
      found_libfuzzer_crash = False
    elif LIBFUZZER_LOG_END_REGEX.match(line):
      libfuzzer_lines_count += 1
      break
    else:
      other_lines_count += 1
      lines_after_last_libfuzzer_line_count += 1

  # Ignore the lines after the last libfuzzer line.
  other_lines_count -= lines_after_last_libfuzzer_line_count
  ignored_lines_count += lines_after_last_libfuzzer_line_count

  return other_lines_count, libfuzzer_lines_count, ignored_lines_count


def strategy_column_name(strategy_name):
  """Convert the strategy name into stats column name."""
  return 'strategy_%s' % strategy_name


def parse_fuzzing_strategies(log_lines, strategies):
  """Extract stats for fuzzing strategies used."""
  if not strategies:
    # Extract strategies from the log.
    for line in log_lines:
      match = LIBFUZZER_FUZZING_STRATEGIES.match(line)
      if match:
        strategies = match.group(1).split(',')
        break

  return process_strategies(strategies)


def process_strategies(strategies, name_modifier=strategy_column_name):
  """Process strategies, parsing any stored values."""
  stats = {}

  def parse_line_for_strategy_prefix(line, strategy_name):
    """Parse log line to find the value of a strategy with a prefix."""
    strategy_prefix = strategy_name + '_'
    if not line.startswith(strategy_prefix):
      return

    suffix_type = strategy.LIBFUZZER_STRATEGIES_WITH_PREFIX_VALUE_TYPE[
        strategy_name]
    try:
      strategy_value = suffix_type(line[len(strategy_prefix):])
      stats[name_modifier(strategy_name)] = strategy_value
    except (IndexError, ValueError) as e:
      logs.log_error('Failed to parse strategy "%s":\n%s\n' % (line, str(e)))

  # These strategies are used with different values specified in the prefix.
  for strategy_type in strategy.LIBFUZZER_STRATEGIES_WITH_PREFIX_VALUE:
    for line in strategies:
      parse_line_for_strategy_prefix(line, strategy_type.name)

  # Other strategies are either ON or OFF, without arbitrary values.
  for strategy_type in strategy.LIBFUZZER_STRATEGIES_WITH_BOOLEAN_VALUE:
    if strategy_type.name in strategies:
      stats[name_modifier(strategy_type.name)] = 1

  return stats


def parse_performance_features(log_lines, strategies, arguments):
  """Extract stats for performance analysis."""
  # TODO(ochang): Remove include_strategies once refactor is complete.
  # Initialize stats with default values.
  stats = {
      'bad_instrumentation': 0,
      'corpus_crash_count': 0,
      'corpus_size': 0,
      'crash_count': 0,
      'dict_used': 0,
      'edge_coverage': 0,
      'edges_total': 0,
      'feature_coverage': 0,
      'initial_edge_coverage': 0,
      'initial_feature_coverage': 0,
      'leak_count': 0,
      'log_lines_unwanted': 0,
      'log_lines_from_engine': 0,
      'log_lines_ignored': 0,
      'max_len': 0,
      'manual_dict_size': 0,
      'merge_edge_coverage': 0,
      'new_edges': 0,
      'new_features': 0,
      'oom_count': 0,
      'recommended_dict_size': 0,
      'slow_unit_count': 0,
      'slow_units_count': 0,
      'startup_crash_count': 1,
      'timeout_count': 0,
  }

  # Extract strategy selection method.
  # TODO(ochang): Move to more general place?
  stats['strategy_selection_method'] = environment.get_value(
      'STRATEGY_SELECTION_METHOD', default_value='default')

  # Initialize all strategy stats as disabled by default.
  for strategy_type in strategy.LIBFUZZER_STRATEGY_LIST:
    if strategy.LIBFUZZER_STRATEGIES_WITH_PREFIX_VALUE_TYPE.get(
        strategy_type.name) == str:
      stats[strategy_column_name(strategy_type.name)] = ''
    else:
      stats[strategy_column_name(strategy_type.name)] = 0

  # Process fuzzing strategies used.
  stats.update(parse_fuzzing_strategies(log_lines, strategies))

  (stats['log_lines_unwanted'], stats['log_lines_from_engine'],
   stats['log_lines_ignored']) = calculate_log_lines(log_lines)

  if stats['log_lines_from_engine'] > 0:
    stats['startup_crash_count'] = 0

  # Extract '-max_len' value from arguments, if possible.
  stats['max_len'] = int(
      fuzzer_utils.extract_argument(
          arguments, constants.MAX_LEN_FLAG, remove=False) or stats['max_len'])

  # Extract sizes of manual and recommended dictionary used for fuzzing.
  dictionary_path = fuzzer_utils.extract_argument(
      arguments, constants.DICT_FLAG, remove=False)
  stats['manual_dict_size'], stats['recommended_dict_size'] = (
      dictionary_manager.get_stats_for_dictionary_file(dictionary_path))

  # Different crashes and other flags extracted via regexp match.
  has_corpus = False
  for line in log_lines:
    if LIBFUZZER_BAD_INSTRUMENTATION_REGEX.match(line):
      stats['bad_instrumentation'] = 1
      continue

    if LIBFUZZER_CRASH_TESTCASE_REGEX.match(line):
      stats['crash_count'] = 1
      continue

    if LIBFUZZER_LOG_DICTIONARY_REGEX.match(line):
      stats['dict_used'] = 1
      continue

    if LEAK_TESTCASE_REGEX.match(line):
      stats['leak_count'] = 1
      continue

    if (LIBFUZZER_OOM_TESTCASE_REGEX.match(line) or
        stacktrace_constants.OUT_OF_MEMORY_REGEX.match(line)):
      stats['oom_count'] = 1
      continue

    if LIBFUZZER_SLOW_UNIT_TESTCASE_REGEX.match(line):
      # Use |slow_unit_count| to track if this run had any slow units at all.
      # and use |slow_units_count| to track the actual number of slow units in
      # this run (used by performance analyzer).
      stats['slow_unit_count'] = 1
      stats['slow_units_count'] += 1
      continue

    match = LIBFUZZER_LOG_SEED_CORPUS_INFO_REGEX.match(line)
    if match:
      has_corpus = True

    match = LIBFUZZER_MODULES_LOADED_REGEX.match(line)
    if match:
      stats['startup_crash_count'] = 0
      stats['edges_total'] = int(match.group(2))

    if (LIBFUZZER_TIMEOUT_TESTCASE_REGEX.match(line) or
        stacktrace_constants.LIBFUZZER_TIMEOUT_REGEX.match(line)):
      stats['timeout_count'] = 1
      continue

    if not stats['max_len']:
      # Get "max_len" value from the log, if it has not been found in arguments.
      match = LIBFUZZER_LOG_MAX_LEN_REGEX.match(line)
      if match:
        stats['max_len'] = int(match.group(1))
        continue

  if has_corpus and not stats['log_lines_from_engine']:
    stats['corpus_crash_count'] = 1

  return stats


def parse_stats_from_merge_log(log_lines):
  """Extract stats from a log produced by libFuzzer run with -merge=1."""
  stats = {
      'edge_coverage': 0,
      'feature_coverage': 0,
  }

  # Reverse the list as an optimization. The line of our interest is the last.
  for line in reversed(log_lines):
    match = LIBFUZZER_MERGE_LOG_STATS_REGEX.match(line)
    if match:
      stats['edge_coverage'] = int(match.group(2))
      stats['feature_coverage'] = int(match.group(1))
      break

  return stats



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzer related helper functions."""

from clusterfuzz._internal.base import modules

modules.fix_module_search_paths()

# Import some functions that may be used by existing fuzzers before
# fuzzer_utils was split up.

from clusterfuzz._internal.fuzzer_utils.tests import create_testcase_list_file
from clusterfuzz._internal.fuzzer_utils.tests import get_testcases
from clusterfuzz._internal.fuzzer_utils.tests import is_locked
from clusterfuzz._internal.fuzzer_utils.tests import is_valid_testcase_file



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for gsutil."""

import os

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.google_cloud_utils import gsutil
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class GSUtilRunnerTest(fake_filesystem_unittest.TestCase):
  """GSUtilRunner tests."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.system.new_process.ProcessRunner.run_and_wait',
    ])

    test_utils.set_up_pyfakefs(self)
    self.gsutil_runner_obj = gsutil.GSUtilRunner()

  def test_rsync_remote_gcs_1(self):
    """Test rsync."""
    self.gsutil_runner_obj.rsync('gs://source_bucket/source_path',
                                 'gs://target_bucket/target_path')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', '-d', 'gs://source_bucket/source_path',
            'gs://target_bucket/target_path'
        ],
        timeout=18000,
        env=mock.ANY)

  def test_rsync_local_gcs_1(self):
    """Test rsync."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/source_bucket')
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.rsync('gs://source_bucket/source_path',
                                 'gs://target_bucket/target_path')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', '-d',
            '/local/source_bucket/objects/source_path',
            '/local/target_bucket/objects/target_path'
        ],
        timeout=18000,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))

  def test_rsync_remote_gcs_2(self):
    """Test rsync."""
    self.gsutil_runner_obj.rsync(
        'gs://source_bucket/source_path',
        'gs://target_bucket/target_path',
        timeout=1337)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', '-d', 'gs://source_bucket/source_path',
            'gs://target_bucket/target_path'
        ],
        timeout=1337,
        env=mock.ANY)

  def test_rsync_local_gcs_2(self):
    """Test rsync."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/source_bucket')
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.rsync(
        'gs://source_bucket/source_path',
        'gs://target_bucket/target_path',
        timeout=1337)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', '-d',
            '/local/source_bucket/objects/source_path',
            '/local/target_bucket/objects/target_path'
        ],
        timeout=1337,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))

  def test_rsync_remote_gcs_3(self):
    """Test rsync."""
    self.gsutil_runner_obj.rsync(
        'gs://source_bucket/source_path',
        'gs://target_bucket/target_path',
        delete=False)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', 'gs://source_bucket/source_path',
            'gs://target_bucket/target_path'
        ],
        timeout=18000,
        env=mock.ANY)

  def test_rsync_local_gcs_3(self):
    """Test rsync."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/source_bucket')
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.rsync(
        'gs://source_bucket/source_path',
        'gs://target_bucket/target_path',
        delete=False)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', '/local/source_bucket/objects/source_path',
            '/local/target_bucket/objects/target_path'
        ],
        timeout=18000,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))

  def test_rsync_remote_gcs_4(self):
    """Test rsync."""
    self.gsutil_runner_obj.rsync(
        'gs://source_bucket/source_path',
        'gs://target_bucket/target_path',
        timeout=1337,
        delete=False)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', 'gs://source_bucket/source_path',
            'gs://target_bucket/target_path'
        ],
        timeout=1337,
        env=mock.ANY)

  def test_rsync_local_gcs_4(self):
    """Test rsync."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/source_bucket')
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.rsync(
        'gs://source_bucket/source_path',
        'gs://target_bucket/target_path',
        timeout=1337,
        delete=False)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', '/local/source_bucket/objects/source_path',
            '/local/target_bucket/objects/target_path'
        ],
        timeout=1337,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))

  def test_rsync_remote_gcs_5(self):
    """Test rsync."""
    self.gsutil_runner_obj.rsync(
        'gs://source_bucket/source_path',
        'gs://target_bucket/target_path',
        timeout=1337,
        delete=False,
        exclusion_pattern='"*.txt$"')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', '-x', '"*.txt$"',
            'gs://source_bucket/source_path', 'gs://target_bucket/target_path'
        ],
        timeout=1337,
        env=mock.ANY)

  def test_rsync_local_gcs_5(self):
    """Test rsync."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/source_bucket')
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.rsync(
        'gs://source_bucket/source_path',
        'gs://target_bucket/target_path',
        timeout=1337,
        delete=False,
        exclusion_pattern='"*.txt$"')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-q', 'rsync', '-r', '-x', '"*.txt$"',
            '/local/source_bucket/objects/source_path',
            '/local/target_bucket/objects/target_path'
        ],
        timeout=1337,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))

  def test_download_file_remote_gcs_1(self):
    """Test download_file."""
    self.gsutil_runner_obj.download_file('gs://source_bucket/source_path',
                                         '/target_path')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', 'gs://source_bucket/source_path', '/target_path'],
        timeout=None,
        env=mock.ANY)

  def test_download_file_local_gcs_1(self):
    """Test download_file."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.gsutil_runner_obj.download_file('gs://source_bucket/source_path',
                                         '/target_path')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', '/local/source_bucket/objects/source_path', '/target_path'],
        timeout=None,
        env=mock.ANY)

  def test_download_file_remote_gcs_2(self):
    """Test download_file."""
    self.gsutil_runner_obj.download_file(
        'gs://source_bucket/source_path', '/target_path', timeout=1337)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', 'gs://source_bucket/source_path', '/target_path'],
        timeout=1337,
        env=mock.ANY)

  def test_download_file_local_gcs_2(self):
    """Test download_file."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.gsutil_runner_obj.download_file(
        'gs://source_bucket/source_path', '/target_path', timeout=1337)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', '/local/source_bucket/objects/source_path', '/target_path'],
        timeout=1337,
        env=mock.ANY)

  def test_upload_file_remote_gcs_1(self):
    """Test upload_file."""
    self.gsutil_runner_obj.upload_file('/source_path',
                                       'gs://target_bucket/target_path')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', '/source_path', 'gs://target_bucket/target_path'],
        timeout=None,
        env=mock.ANY)

  def test_upload_file_local_gcs_1(self):
    """Test upload_file."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.upload_file('/source_path',
                                       'gs://target_bucket/target_path')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', '/source_path', '/local/target_bucket/objects/target_path'],
        timeout=None,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))

  def test_upload_file_remote_gcs_2(self):
    """Test upload_file."""
    self.gsutil_runner_obj.upload_file(
        '/source_path',
        'gs://target_bucket/target_path',
        timeout=1337,
        gzip=True,
        metadata={'a': 'b'})
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-h', 'a:b', 'cp', '-Z', '/source_path',
            'gs://target_bucket/target_path'
        ],
        timeout=1337,
        env=mock.ANY)

  def test_upload_file_local_gcs_2(self):
    """Test upload_file."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.upload_file(
        '/source_path',
        'gs://target_bucket/target_path',
        timeout=1337,
        gzip=True,
        metadata={'a': 'b'})
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner, [
            '-h', 'a:b', 'cp', '-Z', '/source_path',
            '/local/target_bucket/objects/target_path'
        ],
        timeout=1337,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))

  def test_upload_files_to_url_remote_gcs_1(self):
    """Test upload_files_to_url."""
    self.gsutil_runner_obj.upload_files_to_url(
        ['/source_path1', '/source_path2'], 'gs://target_bucket/target_path')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', '-I', 'gs://target_bucket/target_path'],
        input_data=b'/source_path1\n/source_path2',
        timeout=None,
        env=mock.ANY)

  def test_upload_files_local_gcs_1(self):
    """Test upload_files_to_url."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.upload_files_to_url(
        ['/source_path1', '/source_path2'], 'gs://target_bucket/target_path')
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', '-I', '/local/target_bucket/objects/target_path'],
        input_data=b'/source_path1\n/source_path2',
        timeout=None,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))

  def test_upload_files_remote_gcs_2(self):
    """Test upload_files_to_url."""
    self.gsutil_runner_obj.upload_files_to_url(
        ['/source_path1', '/source_path2'],
        'gs://target_bucket/target_path',
        timeout=1337)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', '-I', 'gs://target_bucket/target_path'],
        input_data=b'/source_path1\n/source_path2',
        timeout=1337,
        env=mock.ANY)

  def test_upload_files_to_url_local_gcs_2(self):
    """Test upload_files_to_url."""
    os.environ['LOCAL_GCS_BUCKETS_PATH'] = '/local'
    self.fs.create_dir('/local/target_bucket')
    self.gsutil_runner_obj.upload_files_to_url(
        ['/source_path1', '/source_path2'],
        'gs://target_bucket/target_path',
        timeout=1337)
    self.mock.run_and_wait.assert_called_with(
        self.gsutil_runner_obj.gsutil_runner,
        ['cp', '-I', '/local/target_bucket/objects/target_path'],
        input_data=b'/source_path1\n/source_path2',
        timeout=1337,
        env=mock.ANY)
    self.assertTrue(os.path.exists('/local/target_bucket/objects'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for local_config functions."""

import os
import unittest

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class GetTest(unittest.TestCase):
  """Tests get()."""

  def setUp(self):
    test_helpers.patch_environ(self)

    self.configs_directory = os.path.join(
        os.path.dirname(__file__), 'local_config_data')
    environment.set_value('CONFIG_DIR_OVERRIDE', self.configs_directory)

    self.config = local_config.Config()

  def test_get_with_non_existent_configs_directory(self):
    """Test with non-existent configs directory."""
    environment.set_value('CONFIG_DIR_OVERRIDE', 'non-existent')
    with self.assertRaises(errors.BadConfigError):
      local_config.Config().get('foo')

  def test_get_with_invalid_key_name(self):
    """Test with an invalid key name passed for a lookup."""
    for key in ['', None]:
      with self.assertRaises(errors.InvalidConfigKey):
        self.config.get(key)

  def test_get_with_bad_yaml_file(self):
    """Test with a unparseable yaml file."""
    with self.assertRaises(errors.ConfigParseError):
      self.config.get('bad')

  def test_get_with_root_yaml_file(self):
    """Test with a yaml file in root."""
    self.assertEqual({'b': {'c': 'd', 'e': 1}}, self.config.get('a'))
    self.assertEqual({'c': 'd', 'e': 1}, self.config.get('a.b'))
    self.assertEqual('d', self.config.get('a.b.c'))
    self.assertEqual(1, self.config.get('a.b.e'))
    self.assertEqual(
        os.path.join(self.configs_directory, '1'),
        self.config.get_absolute_path('a.b.e'))

  def test_get_with_subfolder_yaml_file(self):
    """Test with a yaml file in sub-folders."""
    # Use aa/bb/cc.yaml.
    self.assertEqual({'dd': 'ee'}, self.config.get('aa.bb.cc'))
    self.assertEqual('ee', self.config.get('aa.bb.cc.dd'))
    self.assertEqual(
        os.path.join(self.configs_directory, 'aa', 'bb', 'ee'),
        self.config.get_absolute_path('aa.bb.cc.dd'))

    with self.assertRaises(errors.InvalidConfigKey):
      self.config.get('ambiguous.a')

  def test_get_with_sub_config(self):
    """Test with a sub-config."""
    sub_config = local_config.Config().sub_config('aa.bb.cc')
    self.assertEqual('ee', sub_config.get('dd'))
    self.assertEqual({'dd': 'ee'}, sub_config.get())

    sub_config = local_config.Config('aa').sub_config('bb.cc')
    self.assertEqual('ee', sub_config.get('dd'))
    self.assertEqual({'dd': 'ee'}, sub_config.get())

    sub_config = local_config.Config('aa.bb').sub_config('cc')
    self.assertEqual('ee', sub_config.get('dd'))
    self.assertEqual({'dd': 'ee'}, sub_config.get())

  def test_get_with_invalid_keys(self):
    """Test with invalid keys."""
    # Invalid keys, are actually values.
    with self.assertRaises(errors.InvalidConfigKey):
      self.config.get('aa.bb.cc.dd.ee')

  def test_get_non_existent_attributes_without_default(self):
    """Tests non-existent attributes return None without default value."""
    self.assertIsNone(self.config.get('a.b.f'))
    self.assertIsNone(self.config.get('aa.bb.cc.ff'))
    self.assertIsNone(self.config.get('aa.bb.zz'))

    # Does not exist.
    self.assertIsNone(self.config.get('zz.aa'))

  def test_get_non_existent_attributes_with_default(self):
    """Tests non-existent attributes return default with default value."""
    self.assertEqual(2, self.config.get('a.b.f', default=2))
    self.assertEqual(2, self.config.get('aa.bb.cc.ff', default=2))
    self.assertEqual(2, self.config.get('aa.bb.zz', default=2))

    # Does not exist.
    self.assertEqual(2, self.config.get('zz.aa', default=2))

  def test_root_validation(self):
    """Test root validation."""
    _ = local_config.Config()
    _ = local_config.Config('a')
    _ = local_config.Config('aa')
    _ = local_config.Config('aa.bb')
    _ = local_config.Config('aa.bb.cc')
    _ = local_config.Config('aa.bb.cc.dd')

    with self.assertRaises(errors.BadConfigError):
      _ = local_config.Config('aa.b')

    with self.assertRaises(errors.BadConfigError):
      _ = local_config.Config('aa.bb.c')

    with self.assertRaises(errors.BadConfigError):
      _ = local_config.Config('aa.bb.cc.d')

    with self.assertRaises(errors.InvalidConfigKey):
      _ = local_config.Config('aa.bb.cc.dd.ee')


class GetTestWithCache(unittest.TestCase):
  """Tests get() with and without caching."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.config.local_config._search_key',
    ])
    self.mock._search_key.return_value = 'value'  # pylint: disable=protected-access

  def test_with_cache(self):
    """Test that we invoke _search_key once with caching enabled."""
    config = local_config.Config()
    for _ in range(10):
      self.assertEqual('value', config.get('a.b.c'))
    self.assertEqual(1, self.mock._search_key.call_count)  # pylint: disable=protected-access


class ProjectConfigTest(unittest.TestCase):
  """Tests ProjectConfig."""

  def setUp(self):
    test_helpers.patch_environ(self)

    environment.remove_key('PROJECT_NAME')
    environment.remove_key('ISSUE_TRACKER')
    environment.remove_key('UPDATE_WEB_TESTS')

    self.config = local_config.ProjectConfig()

  def test_set_environment_without_default(self):
    """Test that set_environment sets the variable from test config.."""
    self.config.set_environment()
    self.assertEqual('test-project', environment.get_value('PROJECT_NAME'))
    self.assertEqual('test-clusterfuzz',
                     environment.get_value('APPLICATION_ID'))

  def test_set_environment_with_default(self):
    """Test that set_environment sets the variable from test config, skipping
    the ones already set in environment."""
    environment.set_value('ISSUE_TRACKER', 'test-issue-tracker-override')
    environment.set_value('UPDATE_WEB_TESTS', True)
    self.config.set_environment()
    self.assertEqual('test-project', environment.get_value('PROJECT_NAME'))
    self.assertEqual('test-issue-tracker-override',
                     environment.get_value('ISSUE_TRACKER'))
    self.assertEqual(True, environment.get_value('UPDATE_WEB_TESTS'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Builtin fuzzer."""
# NOTE: This module is deprecated and will be replaced with
# lib.clusterfuzz.fuzz.engine.

import os
import random
import sys

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers import utils as fuzzers_utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell


class BuiltinFuzzerResult(object):
  """Result of running a builtin fuzzer."""

  def __init__(self, output, corpus_directory=None):
    self.output = output
    self.corpus_directory = corpus_directory


class BuiltinFuzzerException(Exception):
  """Exception that should be thrown when there is an issue preventing a builtin
  fuzzer from running, or if there is a very unusual exception encountered
  during a run."""


class BuiltinFuzzer(object):
  """Builtin fuzzer."""

  def run(self, input_directory, output_directory, no_of_files):
    raise NotImplementedError

  @property
  def fuzzer_directory(self):
    return os.path.abspath(
        os.path.dirname(sys.modules[self.__module__].__file__))


def get_corpus_directory(input_directory, project_qualified_name):
  """Get the corpus directory given a project qualified fuzz target name."""
  corpus_directory = os.path.join(input_directory, project_qualified_name)
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    corpus_directory = file_host.rebase_to_worker_root(corpus_directory)

  # Create corpus directory if it does not exist already.
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    file_host.create_directory(corpus_directory, create_intermediates=True)
  else:
    shell.create_directory(corpus_directory)
  return corpus_directory


class EngineFuzzer(BuiltinFuzzer):
  """Builtin fuzzer for fuzzing engines such as libFuzzer."""

  def generate_arguments(self, fuzzer_path):
    """Generate arguments for the given fuzzer."""
    raise NotImplementedError

  def _get_fuzzer_binary_name_and_path(self):
    """Returns the fuzzer binary name and its path."""
    # Fuchsia doesn't use file paths to call fuzzers, just the name of the
    # fuzzer, so we set both from FUZZ_TARGET here.
    if environment.platform() == 'FUCHSIA':
      fuzzer_binary_name = fuzzer_path = environment.get_value('FUZZ_TARGET')
      return fuzzer_binary_name, fuzzer_path
    build_directory = environment.get_value('BUILD_DIR')

    if not build_directory:
      raise BuiltinFuzzerException('BUILD_DIR environment variable is not set.')

    fuzzers = fuzzers_utils.get_fuzz_targets(build_directory)

    if not fuzzers:
      raise BuiltinFuzzerException(
          'No fuzzer binaries found in |BUILD_DIR| directory.')

    fuzzer_binary_name = environment.get_value('FUZZ_TARGET')
    if fuzzer_binary_name:
      fuzzer_path = _get_fuzzer_path(fuzzers, fuzzer_binary_name)
    else:
      fuzzer_path = random.SystemRandom().choice(fuzzers)
      fuzzer_binary_name = os.path.basename(fuzzer_path)
    return fuzzer_binary_name, fuzzer_path

  def run(self, input_directory, output_directory, no_of_files):
    """Run the fuzzer to generate testcases."""

    fuzzer_binary_name, fuzzer_path = self._get_fuzzer_binary_name_and_path()

    project_qualified_name = data_types.fuzz_target_project_qualified_name(
        utils.current_project(), fuzzer_binary_name)

    arguments = self.generate_arguments(fuzzer_path)
    corpus_directory = get_corpus_directory(input_directory,
                                            project_qualified_name)

    # Create fuzz testcases.
    for i in range(no_of_files):
      # Contents of testcase file don't matter at this point. Need to create
      # something non-null so that it is not ignored.
      testcase_file_path = os.path.join(
          output_directory, '%s%d' % (testcase_manager.FUZZ_PREFIX, i))
      utils.write_data_to_file(' ', testcase_file_path)

      # Write the flags file containing command line for running launcher
      # script.
      flags_file_path = os.path.join(
          output_directory, '%s%d' % (testcase_manager.FLAGS_PREFIX, i))
      flags = ['%TESTCASE%', fuzzer_binary_name]
      if arguments:
        flags.append(arguments)

      flags_file_content = ' '.join(flags)
      utils.write_data_to_file(flags_file_content, flags_file_path)

    output = 'Generated %d testcase for fuzzer %s.\n' % (no_of_files,
                                                         fuzzer_binary_name)
    output += 'metadata::fuzzer_binary_name: %s\n' % fuzzer_binary_name

    issue_owners = engine_common.get_issue_owners(fuzzer_path)
    if issue_owners:
      output += 'metadata::issue_owners: %s\n' % ','.join(issue_owners)

    issue_labels = engine_common.get_issue_labels(fuzzer_path)
    if issue_labels:
      output += 'metadata::issue_labels: %s\n' % ','.join(issue_labels)

    issue_components = engine_common.get_issue_components(fuzzer_path)
    if issue_components:
      output += 'metadata::issue_components: %s\n' % ','.join(issue_components)

    # Update *SAN_OPTIONS in current environment from .options file. This
    # environment is used in fuzz task later for deriving the environment
    # string in |get_environment_settings_as_string| and embedding this as
    # part of stacktrace.
    engine_common.process_sanitizer_options_overrides(fuzzer_path)

    return BuiltinFuzzerResult(output=output, corpus_directory=corpus_directory)


def _get_fuzzer_path(target_list, fuzzer_name):
  """Return the full fuzzer path and actual binary name of |fuzzer_name|."""
  fuzzer_filename = environment.get_executable_filename(fuzzer_name)
  for path in target_list:
    if os.path.basename(path) == fuzzer_filename:
      return path

  raise BuiltinFuzzerException('Failed to find chosen target ' + fuzzer_name)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Blobs handling."""

import os
import re
import uuid

from google.cloud import ndb

from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import retry
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment

from . import storage

FAIL_NUM_RETRIES = 2
FAIL_WAIT = 1.5


# pylint: disable=invalid-name
class BlobInfo(data_types.Model):
  """Legacy BlobInfo."""

  content_type = ndb.StringProperty()
  creation = ndb.DateTimeProperty()
  filename = ndb.StringProperty()
  gs_object_name = ndb.StringProperty()
  md5_hash = ndb.StringProperty()
  size = ndb.IntegerProperty()
  upload_id = ndb.StringProperty()

  @classmethod
  def _get_kind(cls):
    if environment.get_value('DATASTORE_EMULATOR_HOST'):
      # Datastore emulator does not allow writing entities with names of the
      # format "__*__".
      cls._kind_map['_BlobInfo_'] = cls
      return '_BlobInfo_'

    return '__BlobInfo__'


class _blobmigrator_BlobKeyMapping(data_types.Model):
  """Migrated blob."""

  old_blob_key = ndb.ComputedProperty(lambda self: self.key.id())
  gcs_filename = ndb.StringProperty(required=True)
  new_blob_key = ndb.StringProperty(required=True)


# pylint: enable=invalid-name
class BlobsException(Exception):
  """Base exception for blobs module."""


def _is_gcs_key(blob_key):
  """Return whether if the key is a GCS key."""
  gcs_key_pattern = re.compile(
      r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$')

  return bool(gcs_key_pattern.match(blob_key))


def _get_gcs_blob_path(blob_key):
  """Return the full path to the blob on GCS."""
  return '/%s/%s' % (storage.blobs_bucket(), blob_key)


def get_gcs_path(blob_key):
  """Return GCS path for the blob."""
  if _is_gcs_key(blob_key):
    return _get_gcs_blob_path(blob_key)

  # Legacy blobstore key.
  blob_info = get_legacy_blob_info(blob_key)
  if not blob_info:
    return None

  return blob_info.gs_object_name


@memoize.wrap(memoize.Memcache(60 * 60 * 24 * 30))  # 30 day TTL
@retry.wrap(
    retries=FAIL_NUM_RETRIES,
    delay=FAIL_WAIT,
    function='google_cloud_utils.blobs.get_blob_size')
def get_blob_size(blob_key):
  """Returns blob size for a given blob key."""
  if not blob_key or blob_key == 'NA':
    return None

  blob_info = get_blob_info(blob_key)
  if not blob_info:
    return None

  return blob_info.size


def get_blob_info(blob_key):
  """Get the GcsBlobInfo for the given key. Always returns a
  storage.GcsBlobInfo, even for legacy blobs."""
  if _is_gcs_key(blob_key):
    return storage.GcsBlobInfo.from_key(blob_key)

  legacy_blob_info = get_legacy_blob_info(blob_key)
  if not legacy_blob_info:
    return None

  return storage.GcsBlobInfo.from_legacy_blob_info(legacy_blob_info)


@retry.wrap(
    retries=FAIL_NUM_RETRIES,
    delay=FAIL_WAIT,
    function='google_cloud_utils.blobs.delete_blob',
    retry_on_false=True)
def delete_blob(blob_key):
  """Delete a blob key."""
  blob_info = get_blob_info(blob_key)
  if not blob_info:
    return False

  return storage.delete(blob_info.gcs_path)


@retry.wrap(
    retries=FAIL_NUM_RETRIES,
    delay=FAIL_WAIT,
    function='google_cloud_utils.blobs.write_blob',
    retry_on_false=True)
def write_blob(file_handle_or_path):
  """Write a single file testcase to GCS."""
  blobs_bucket = storage.blobs_bucket()
  blob_name = generate_new_blob_name()

  if storage.get(storage.get_cloud_storage_file_path(blobs_bucket, blob_name)):
    raise BlobsException('UUID collision found: %s' % blob_name)

  if isinstance(file_handle_or_path, str):
    filename = os.path.basename(file_handle_or_path)
  else:
    filename = file_handle_or_path.name

  metadata = {
      storage.BLOB_FILENAME_METADATA_KEY: filename,
  }

  gcs_path = '/%s/%s' % (blobs_bucket, blob_name)
  if storage.copy_file_to(file_handle_or_path, gcs_path, metadata=metadata):
    return blob_name

  raise BlobsException('Failed to write blob %s.' % blob_name)


@retry.wrap(
    retries=FAIL_NUM_RETRIES,
    delay=FAIL_WAIT,
    function='google_cloud_utils.blobs.read_blob_to_disk',
    retry_on_false=True)
def read_blob_to_disk(blob_key, local_file):
  """Copy data stored in the blobstore to a local file."""
  assert not environment.is_running_on_app_engine()

  directory = os.path.dirname(local_file)
  if not os.path.exists(directory):
    os.makedirs(directory)

  gcs_path = get_gcs_path(blob_key)
  return storage.copy_file_from(gcs_path, local_file)


def read_key(blob_key):
  """Returns data associated with a blobstore key."""
  gcs_path = get_gcs_path(blob_key)
  return storage.read_data(gcs_path)


def get_legacy_blob_info(blob_key):
  """Return legacy blob info information."""
  legacy_blob_info = ndb.Key(BlobInfo, blob_key).get()
  if not legacy_blob_info:
    return None

  if legacy_blob_info.gs_object_name:
    return legacy_blob_info

  # Blobs which were stored before the move to GCS have an additional mapping
  # entry created by our migration jobs.
  blob_mapping = get_blob_mapping(blob_key)
  if not blob_mapping:
    raise BlobsException('Blob mapping not found.')

  legacy_blob_info.gs_object_name = blob_mapping.gcs_filename
  return legacy_blob_info


def get_blob_mapping(blob_key):
  """Return blob mapping information."""
  return ndb.Key(_blobmigrator_BlobKeyMapping, blob_key).get()


def generate_new_blob_name():
  """Generate a new blob name."""
  return str(uuid.uuid4()).lower()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fuzzer.py."""

import os
import unittest

from clusterfuzz._internal.bot.fuzzers.libFuzzer import fuzzer
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.core.bot.fuzzers import builtin_test
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


class GenerateArgumentsTests(unittest.TestCase):
  """Unit tests for fuzzer.py."""

  def setUp(self):
    """Set up test environment."""
    test_helpers.patch_environ(self)
    environment.set_value('FUZZ_TEST_TIMEOUT', '4800')

    self.build_dir = os.path.join(SCRIPT_DIR, 'run_data', 'build_dir')
    self.corpus_directory = 'data/corpus_with_some_files'

  def test_generate_arguments_default(self):
    """Test generateArgumentsForFuzzer."""
    fuzzer_path = os.path.join(self.build_dir, 'fake0_fuzzer')
    libfuzzer = fuzzer.LibFuzzer()
    arguments = libfuzzer.generate_arguments(fuzzer_path)
    expected_arguments = '-timeout=25 -rss_limit_mb=2560'

    self.assertEqual(arguments, expected_arguments)

  def test_generate_arguments_with_options_file(self):
    """Test generateArgumentsForFuzzer."""
    fuzzer_path = os.path.join(self.build_dir, 'fake1_fuzzer')
    libfuzzer = fuzzer.LibFuzzer()
    arguments = libfuzzer.generate_arguments(fuzzer_path)

    expected_arguments = (
        '-max_len=31337 -timeout=11 -runs=9999999 -rss_limit_mb=2560')
    self.assertEqual(arguments, expected_arguments)


class FuzzerTest(builtin_test.BaseEngineFuzzerTest):
  """Unit tests for fuzzer."""

  def test_run(self):
    """Test running libFuzzer fuzzer."""
    libfuzzer = fuzzer.LibFuzzer()
    libfuzzer.run('/input', '/output', 1)
    with open('/output/flags-0') as f:
      self.assertEqual('%TESTCASE% target -timeout=25 -rss_limit_mb=2560',
                       f.read())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common helper functions for setup at the start of tasks."""

import datetime
import os
import shlex
import time
import zipfile

import six

from clusterfuzz._internal.base import dates
from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers import builtin_fuzzers
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import locks
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.fuzzing import leak_blacklist
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import fuzzer_logs
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms import android
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

_BOT_DIR = 'bot'
_DATA_BUNDLE_CACHE_COUNT = 10
_DATA_BUNDLE_SYNC_INTERVAL_IN_SECONDS = 6 * 60 * 60
_DATA_BUNDLE_LOCK_INTERVAL_IN_SECONDS = 3 * 60 * 60
_SYNC_FILENAME = '.sync'
_TESTCASE_ARCHIVE_EXTENSION = '.zip'


def _set_timeout_value_from_user_upload(testcase_id):
  """Get the timeout associated with this testcase."""
  metadata = data_types.TestcaseUploadMetadata.query(
      data_types.TestcaseUploadMetadata.testcase_id == int(testcase_id)).get()
  if metadata and metadata.timeout:
    environment.set_value('TEST_TIMEOUT', metadata.timeout)


def _copy_testcase_to_device_and_setup_environment(testcase,
                                                   testcase_file_path):
  """Android specific setup steps for testcase."""
  # Copy test(s) to device.
  android.device.push_testcases_to_device()

  # The following steps need privileged job access.
  job_type_has_privileged_access = environment.get_value('PRIVILEGED_ACCESS')
  if not job_type_has_privileged_access:
    return

  # Install testcase if it is an app.
  package_name = android.app.get_package_name(testcase_file_path)
  if package_name:
    # Set the package name for later use.
    environment.set_value('PKG_NAME', package_name)

    # Install the application apk.
    android.device.install_application_if_needed(
        testcase_file_path, force_update=True)

  # Set app launch command if available from upload.
  app_launch_command = testcase.get_metadata('app_launch_command')
  if app_launch_command:
    environment.set_value('APP_LAUNCH_COMMAND', app_launch_command)

  # Set executable bit on the testcase (to allow binary executable testcases
  # to work in app launch command, e.g. shell %TESTCASE%).
  local_testcases_directory = environment.get_value('FUZZ_INPUTS')
  if (testcase_file_path and
      testcase_file_path.startswith(local_testcases_directory)):
    relative_testcase_file_path = (
        testcase_file_path[len(local_testcases_directory) + 1:])
    device_testcase_file_path = os.path.join(
        android.constants.DEVICE_TESTCASES_DIR, relative_testcase_file_path)
    android.adb.run_shell_command(['chmod', '0755', device_testcase_file_path])


def _get_application_arguments(testcase, job_type, task_name):
  """Get application arguments to use for setting up |testcase|. Use minimized
   arguments if available. For variant task, where we run a testcase against
   another job type, use both minimized arguments and application arguments
   from job."""
  testcase_args = testcase.minimized_arguments
  if not testcase_args:
    return None

  if task_name != 'variant':
    return testcase_args

  # TODO(aarya): Use %TESTCASE% explicitly since it will not exist with new
  # engine impl libFuzzer testcases and AFL's launcher.py requires it as the
  # first argument. Remove once AFL is migrated to the new engine impl.
  if environment.is_afl_job(job_type):
    return '%TESTCASE%'

  job_args = data_handler.get_value_from_job_definition(
      job_type, 'APP_ARGS', default='')
  job_args_list = shlex.split(job_args)
  testcase_args_list = shlex.split(testcase_args)
  testcase_args_filtered_list = [
      arg for arg in testcase_args_list if arg not in job_args_list
  ]

  app_args = ' '.join(testcase_args_filtered_list)
  if job_args:
    if app_args:
      app_args += ' '
    app_args += job_args

  return app_args


def _setup_memory_tools_environment(testcase):
  """Set up environment for various memory tools used."""
  env = testcase.get_metadata('env')
  if not env:
    environment.reset_current_memory_tool_options(
        redzone_size=testcase.redzone, disable_ubsan=testcase.disable_ubsan)
    return

  for options_name, options_value in six.iteritems(env):
    if not options_value:
      environment.remove_key(options_name)
      continue
    environment.set_memory_tool_options(options_name, options_value)


def prepare_environment_for_testcase(testcase, job_type, task_name):
  """Set various environment variables based on the test case."""
  _setup_memory_tools_environment(testcase)

  # Setup environment variable for windows size and location properties.
  # Explicit override to avoid using the default one from job definition since
  # that contains unsubsituted vars like $WIDTH, etc.
  environment.set_value('WINDOW_ARG', testcase.window_argument)

  # Adjust timeout based on the stored multiplier (if available).
  if testcase.timeout_multiplier:
    test_timeout = environment.get_value('TEST_TIMEOUT')
    environment.set_value('TEST_TIMEOUT',
                          int(test_timeout * testcase.timeout_multiplier))

  # Add FUZZ_TARGET to environment if this is a fuzz target testcase.
  fuzz_target = testcase.get_metadata('fuzzer_binary_name')
  if fuzz_target:
    environment.set_value('FUZZ_TARGET', fuzz_target)

  # Override APP_ARGS with minimized arguments (if available). Don't do this
  # for variant task since other job types can have its own set of required
  # arguments, so use the full set of arguments of that job.
  app_args = _get_application_arguments(testcase, job_type, task_name)
  if app_args:
    environment.set_value('APP_ARGS', app_args)


def setup_testcase(testcase, job_type, fuzzer_override=None):
  """Sets up the testcase and needed dependencies like fuzzer,
  data bundle, etc."""
  fuzzer_name = fuzzer_override or testcase.fuzzer_name
  task_name = environment.get_value('TASK_NAME')
  testcase_fail_wait = environment.get_value('FAIL_WAIT')
  testcase_id = testcase.key.id()

  # Clear testcase directories.
  shell.clear_testcase_directories()

  # Adjust the test timeout value if this is coming from an user uploaded
  # testcase.
  if testcase.uploader_email:
    _set_timeout_value_from_user_upload(testcase_id)

  # Update the fuzzer if necessary in order to get the updated data bundle.
  if fuzzer_name:
    try:
      update_successful = update_fuzzer_and_data_bundles(fuzzer_name)
    except errors.InvalidFuzzerError:
      # Close testcase and don't recreate tasks if this fuzzer is invalid.
      testcase.open = False
      testcase.fixed = 'NA'
      testcase.set_metadata('fuzzer_was_deleted', True)
      logs.log_error('Closed testcase %d with invalid fuzzer %s.' %
                     (testcase_id, fuzzer_name))

      error_message = 'Fuzzer %s no longer exists' % fuzzer_name
      data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                           error_message)
      return None, None, None

    if not update_successful:
      error_message = 'Unable to setup fuzzer %s' % fuzzer_name
      data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                           error_message)
      tasks.add_task(
          task_name, testcase_id, job_type, wait_time=testcase_fail_wait)
      return None, None, None

  # Extract the testcase and any of its resources to the input directory.
  file_list, input_directory, testcase_file_path = unpack_testcase(testcase)
  if not file_list:
    error_message = 'Unable to setup testcase %s' % testcase_file_path
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         error_message)
    tasks.add_task(
        task_name, testcase_id, job_type, wait_time=testcase_fail_wait)
    return None, None, None

  # For Android/Fuchsia, we need to sync our local testcases directory with the
  # one on the device.
  if environment.is_android():
    _copy_testcase_to_device_and_setup_environment(testcase, testcase_file_path)

  # Push testcases to worker.
  if environment.is_trusted_host():
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    file_host.push_testcases_to_worker()

  # Copy global blacklist into local blacklist.
  is_lsan_enabled = environment.get_value('LSAN')
  if is_lsan_enabled:
    # Get local blacklist without this testcase's entry.
    leak_blacklist.copy_global_to_local_blacklist(excluded_testcase=testcase)

  prepare_environment_for_testcase(testcase, job_type, task_name)

  return file_list, input_directory, testcase_file_path


def _get_testcase_file_and_path(testcase):
  """Figure out the relative path and input directory for this testcase."""
  testcase_absolute_path = testcase.absolute_path

  # This hack is needed so that we can run a testcase generated on windows, on
  # linux. os.path.isabs return false on paths like c:\a\b\c.
  testcase_path_is_absolute = (
      testcase_absolute_path[1:3] == ':\\' or
      os.path.isabs(testcase_absolute_path))

  # Fix os.sep in testcase path if we are running this on non-windows platform.
  # It is unusual to have '\\' on linux paths, so substitution should be safe.
  if environment.platform() != 'WINDOWS' and '\\' in testcase_absolute_path:
    testcase_absolute_path = testcase_absolute_path.replace('\\', os.sep)

  # Default directory for testcases.
  input_directory = environment.get_value('FUZZ_INPUTS')
  if not testcase_path_is_absolute:
    testcase_path = os.path.join(input_directory, testcase_absolute_path)
    return input_directory, testcase_path

  # Check if the testcase is on a nfs data bundle. If yes, then just
  # return it without doing root directory path fix.
  nfs_root = environment.get_value('NFS_ROOT')
  if nfs_root and testcase_absolute_path.startswith(nfs_root):
    return input_directory, testcase_absolute_path

  # Root directory can be different on bots. Fix the path to account for this.
  root_directory = environment.get_value('ROOT_DIR')
  search_string = '%s%s%s' % (os.sep, _BOT_DIR, os.sep)
  search_index = testcase_absolute_path.find(search_string)
  relative_path = testcase_absolute_path[search_index + len(search_string):]
  testcase_path = os.path.join(root_directory, _BOT_DIR, relative_path)

  return input_directory, testcase_path


def unpack_testcase(testcase):
  """Unpack a testcase and return all files it is composed of."""
  # Figure out where the testcase file should be stored.
  input_directory, testcase_file_path = _get_testcase_file_and_path(testcase)

  minimized = testcase.minimized_keys and testcase.minimized_keys != 'NA'
  if minimized:
    key = testcase.minimized_keys
    archived = bool(testcase.archive_state & data_types.ArchiveStatus.MINIMIZED)
  else:
    key = testcase.fuzzed_keys
    archived = bool(testcase.archive_state & data_types.ArchiveStatus.FUZZED)

  if archived:
    if minimized:
      temp_filename = (
          os.path.join(input_directory,
                       str(testcase.key.id()) + _TESTCASE_ARCHIVE_EXTENSION))
    else:
      temp_filename = os.path.join(input_directory, testcase.archive_filename)
  else:
    temp_filename = testcase_file_path

  if not blobs.read_blob_to_disk(key, temp_filename):
    return None, input_directory, testcase_file_path

  file_list = []
  if archived:
    archive.unpack(temp_filename, input_directory)
    file_list = archive.get_file_list(temp_filename)
    shell.remove_file(temp_filename)

    file_exists = False
    for file_name in file_list:
      if os.path.basename(file_name) == os.path.basename(testcase_file_path):
        file_exists = True
        break

    if not file_exists:
      logs.log_error(
          'Expected file to run %s is not in archive. Base directory is %s and '
          'files in archive are [%s].' % (testcase_file_path, input_directory,
                                          ','.join(file_list)))
      return None, input_directory, testcase_file_path
  else:
    file_list.append(testcase_file_path)

  return file_list, input_directory, testcase_file_path


def _get_data_bundle_update_lock_name(data_bundle_name):
  """Return the lock key name for the given data bundle."""
  return 'update:data_bundle:%s' % data_bundle_name


def _get_data_bundle_sync_file_path(data_bundle_directory):
  """Return path to data bundle sync file."""
  return os.path.join(data_bundle_directory, _SYNC_FILENAME)


def _fetch_lock_for_data_bundle_update(data_bundle):
  """Fetch exclusive lock for a data bundle update."""
  # Data bundle on local disk can be modified without race conditions.
  if data_bundle.is_local:
    return True

  data_bundle_lock_name = _get_data_bundle_update_lock_name(data_bundle.name)
  return locks.acquire_lock(
      data_bundle_lock_name,
      max_hold_seconds=_DATA_BUNDLE_LOCK_INTERVAL_IN_SECONDS,
      by_zone=True)


def _clear_old_data_bundles_if_needed():
  """Clear old data bundles so as to keep the disk cache restricted to
  |_DATA_BUNDLE_CACHE_COUNT| data bundles and prevent potential out-of-disk
  spaces."""
  data_bundles_directory = environment.get_value('DATA_BUNDLES_DIR')

  dirs = []
  for filename in os.listdir(data_bundles_directory):
    file_path = os.path.join(data_bundles_directory, filename)
    if not os.path.isdir(file_path):
      continue
    dirs.append(file_path)

  dirs_to_remove = sorted(
      dirs, key=os.path.getmtime, reverse=True)[_DATA_BUNDLE_CACHE_COUNT:]
  for dir_to_remove in dirs_to_remove:
    logs.log('Removing data bundle directory to keep disk cache small: %s' %
             dir_to_remove)
    shell.remove_directory(dir_to_remove)


def update_data_bundle(fuzzer, data_bundle):
  """Updates a data bundle to the latest version."""
  # This module can't be in the global imports due to appengine issues
  # with multiprocessing and psutil imports.
  from clusterfuzz._internal.google_cloud_utils import gsutil

  # If we are using a data bundle on NFS, it is expected that our testcases
  # will usually be large enough that we would fill up our tmpfs directory
  # pretty quickly. So, change it to use an on-disk directory.
  if not data_bundle.is_local:
    testcase_disk_directory = environment.get_value('FUZZ_INPUTS_DISK')
    environment.set_value('FUZZ_INPUTS', testcase_disk_directory)

  data_bundle_directory = get_data_bundle_directory(fuzzer.name)
  if not data_bundle_directory:
    logs.log_error('Failed to setup data bundle %s.' % data_bundle.name)
    return False

  if not shell.create_directory(
      data_bundle_directory, create_intermediates=True):
    logs.log_error(
        'Failed to create data bundle %s directory.' % data_bundle.name)
    return False

  # Check if data bundle is up to date. If yes, skip the update.
  if _is_data_bundle_up_to_date(data_bundle, data_bundle_directory):
    logs.log('Data bundle was recently synced, skip.')
    return True

  # Fetch lock for this data bundle.
  if not _fetch_lock_for_data_bundle_update(data_bundle):
    logs.log_error('Failed to lock data bundle %s.' % data_bundle.name)
    return False

  # Re-check if another bot did the sync already. If yes, skip.
  if _is_data_bundle_up_to_date(data_bundle, data_bundle_directory):
    logs.log('Another bot finished the sync, skip.')
    _release_lock_for_data_bundle_update(data_bundle)
    return True

  time_before_sync_start = time.time()

  # No need to sync anything if this is a search index data bundle. In that
  # case, the fuzzer will generate testcases from a gcs bucket periodically.
  if not _is_search_index_data_bundle(data_bundle.name):
    bucket_url = data_handler.get_data_bundle_bucket_url(data_bundle.name)

    if environment.is_trusted_host() and data_bundle.sync_to_worker:
      from clusterfuzz._internal.bot.untrusted_runner import corpus_manager
      from clusterfuzz._internal.bot.untrusted_runner import file_host
      worker_data_bundle_directory = file_host.rebase_to_worker_root(
          data_bundle_directory)

      file_host.create_directory(
          worker_data_bundle_directory, create_intermediates=True)
      result = corpus_manager.RemoteGSUtilRunner().rsync(
          bucket_url, worker_data_bundle_directory, delete=False)
    else:
      result = gsutil.GSUtilRunner().rsync(
          bucket_url, data_bundle_directory, delete=False)

    if result.return_code != 0:
      logs.log_error('Failed to sync data bundle %s: %s.' % (data_bundle.name,
                                                             result.output))
      _release_lock_for_data_bundle_update(data_bundle)
      return False

  # Update the testcase list file.
  testcase_manager.create_testcase_list_file(data_bundle_directory)

  #  Write last synced time in the sync file.
  sync_file_path = _get_data_bundle_sync_file_path(data_bundle_directory)
  utils.write_data_to_file(time_before_sync_start, sync_file_path)
  if environment.is_trusted_host() and data_bundle.sync_to_worker:
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    worker_sync_file_path = file_host.rebase_to_worker_root(sync_file_path)
    file_host.copy_file_to_worker(sync_file_path, worker_sync_file_path)

  # Release acquired lock.
  _release_lock_for_data_bundle_update(data_bundle)

  return True


def update_fuzzer_and_data_bundles(fuzzer_name):
  """Update the fuzzer with a given name if necessary."""
  fuzzer = data_types.Fuzzer.query(data_types.Fuzzer.name == fuzzer_name).get()
  if not fuzzer:
    logs.log_error('No fuzzer exists with name %s.' % fuzzer_name)
    raise errors.InvalidFuzzerError

  # Set some helper environment variables.
  fuzzer_directory = get_fuzzer_directory(fuzzer_name)
  environment.set_value('FUZZER_DIR', fuzzer_directory)
  environment.set_value('UNTRUSTED_CONTENT', fuzzer.untrusted_content)

  # If the fuzzer generates large testcases or a large number of small ones
  # that don't fit on tmpfs, then use the larger disk directory.
  if fuzzer.has_large_testcases:
    testcase_disk_directory = environment.get_value('FUZZ_INPUTS_DISK')
    environment.set_value('FUZZ_INPUTS', testcase_disk_directory)

  # Adjust the test timeout, if user has provided one.
  if fuzzer.timeout:
    environment.set_value('TEST_TIMEOUT', fuzzer.timeout)

    # Increase fuzz test timeout if the fuzzer timeout is higher than its
    # current value.
    fuzz_test_timeout = environment.get_value('FUZZ_TEST_TIMEOUT')
    if fuzz_test_timeout and fuzz_test_timeout < fuzzer.timeout:
      environment.set_value('FUZZ_TEST_TIMEOUT', fuzzer.timeout)

  # Adjust the max testcases if this fuzzer has specified a lower limit.
  max_testcases = environment.get_value('MAX_TESTCASES')
  if fuzzer.max_testcases and fuzzer.max_testcases < max_testcases:
    environment.set_value('MAX_TESTCASES', fuzzer.max_testcases)

  # Check for updates to this fuzzer.
  version_file = os.path.join(fuzzer_directory, '.%s_version' % fuzzer_name)
  if (not fuzzer.builtin and
      revisions.needs_update(version_file, fuzzer.revision)):
    logs.log('Fuzzer update was found, updating.')

    # Clear the old fuzzer directory if it exists.
    if not shell.remove_directory(fuzzer_directory, recreate=True):
      logs.log_error('Failed to clear fuzzer directory.')
      return None

    # Copy the archive to local disk and unpack it.
    archive_path = os.path.join(fuzzer_directory, fuzzer.filename)
    if not blobs.read_blob_to_disk(fuzzer.blobstore_key, archive_path):
      logs.log_error('Failed to copy fuzzer archive.')
      return None

    try:
      archive.unpack(archive_path, fuzzer_directory)
    except Exception:
      error_message = ('Failed to unpack fuzzer archive %s '
                       '(bad archive or unsupported format).') % fuzzer.filename
      logs.log_error(error_message)
      fuzzer_logs.upload_script_log(
          'Fatal error: ' + error_message, fuzzer_name=fuzzer_name)
      return None

    fuzzer_path = os.path.join(fuzzer_directory, fuzzer.executable_path)
    if not os.path.exists(fuzzer_path):
      error_message = ('Fuzzer executable %s not found. '
                       'Check fuzzer configuration.') % fuzzer.executable_path
      logs.log_error(error_message)
      fuzzer_logs.upload_script_log(
          'Fatal error: ' + error_message, fuzzer_name=fuzzer_name)
      return None

    # Make fuzzer executable.
    os.chmod(fuzzer_path, 0o750)

    # Cleanup unneeded archive.
    shell.remove_file(archive_path)

    # Save the current revision of this fuzzer in a file for later checks.
    revisions.write_revision_to_revision_file(version_file, fuzzer.revision)
    logs.log('Updated fuzzer to revision %d.' % fuzzer.revision)

  _clear_old_data_bundles_if_needed()

  # Setup data bundles associated with this fuzzer.
  data_bundles = ndb_utils.get_all_from_query(
      data_types.DataBundle.query(
          data_types.DataBundle.name == fuzzer.data_bundle_name))
  for data_bundle in data_bundles:
    if not update_data_bundle(fuzzer, data_bundle):
      return None

  # Setup environment variable for launcher script path.
  if fuzzer.launcher_script:
    fuzzer_launcher_path = os.path.join(fuzzer_directory,
                                        fuzzer.launcher_script)
    environment.set_value('LAUNCHER_PATH', fuzzer_launcher_path)

    # For launcher script usecase, we need the entire fuzzer directory on the
    # worker.
    if environment.is_trusted_host():
      from clusterfuzz._internal.bot.untrusted_runner import file_host
      worker_fuzzer_directory = file_host.rebase_to_worker_root(
          fuzzer_directory)
      file_host.copy_directory_to_worker(
          fuzzer_directory, worker_fuzzer_directory, replace=True)

  return fuzzer


def _is_search_index_data_bundle(data_bundle_name):
  """Return true on if this is a search index data bundle, false otherwise."""
  return data_bundle_name.startswith(
      testcase_manager.SEARCH_INDEX_BUNDLE_PREFIX)


def _is_data_bundle_up_to_date(data_bundle, data_bundle_directory):
  """Return true if the data bundle is up to date, false otherwise."""
  sync_file_path = _get_data_bundle_sync_file_path(data_bundle_directory)

  if environment.is_trusted_host() and data_bundle.sync_to_worker:
    from clusterfuzz._internal.bot.untrusted_runner import file_host
    worker_sync_file_path = file_host.rebase_to_worker_root(sync_file_path)
    shell.remove_file(sync_file_path)
    file_host.copy_file_from_worker(worker_sync_file_path, sync_file_path)

  if not os.path.exists(sync_file_path):
    return False

  last_sync_time = datetime.datetime.utcfromtimestamp(
      utils.read_data_from_file(sync_file_path))

  # Check if we recently synced.
  if not dates.time_has_expired(
      last_sync_time, seconds=_DATA_BUNDLE_SYNC_INTERVAL_IN_SECONDS):
    return True

  # For search index data bundle, we don't sync them from bucket. Instead, we
  # rely on the fuzzer to generate testcases periodically.
  if _is_search_index_data_bundle(data_bundle.name):
    return False

  # Check when the bucket url had last updates. If no new updates, no need to
  # update directory.
  bucket_url = data_handler.get_data_bundle_bucket_url(data_bundle.name)
  last_updated_time = storage.last_updated(bucket_url)
  if last_updated_time and last_sync_time > last_updated_time:
    logs.log(
        'Data bundle %s has no new content from last sync.' % data_bundle.name)
    return True

  return False


def _get_nfs_data_bundle_path(data_bundle_name):
  """Get  path for a data bundle on NFS."""
  nfs_root = environment.get_value('NFS_ROOT')

  # Special naming and path for search index based bundles.
  if _is_search_index_data_bundle(data_bundle_name):
    return os.path.join(
        nfs_root, testcase_manager.SEARCH_INDEX_TESTCASES_DIRNAME,
        data_bundle_name[len(testcase_manager.SEARCH_INDEX_BUNDLE_PREFIX):])

  return os.path.join(nfs_root, data_bundle_name)


def _release_lock_for_data_bundle_update(data_bundle):
  """Release the lock held for the data bundle update."""
  # Data bundle on local disk is never locked in first place.
  if data_bundle.is_local:
    return

  data_bundle_lock_name = _get_data_bundle_update_lock_name(data_bundle.name)
  locks.release_lock(data_bundle_lock_name, by_zone=True)


def get_data_bundle_directory(fuzzer_name):
  """Return data bundle data directory."""
  fuzzer = data_types.Fuzzer.query(data_types.Fuzzer.name == fuzzer_name).get()
  if not fuzzer:
    logs.log_error('Unable to find fuzzer %s.' % fuzzer_name)
    return None

  # Store corpora for built-in fuzzers like libFuzzer in the same directory
  # as other local data bundles. This makes it easy to clear them when we run
  # out of disk space.
  local_data_bundles_directory = environment.get_value('DATA_BUNDLES_DIR')
  if fuzzer.builtin:
    return local_data_bundles_directory

  # Check if we have a fuzzer-specific data bundle. Use it to calculate the
  # data directory we will fetch our testcases from.
  data_bundle = data_types.DataBundle.query(
      data_types.DataBundle.name == fuzzer.data_bundle_name).get()
  if not data_bundle:
    # Generic data bundle directory. Available to all fuzzers if they don't
    # have their own data bundle.
    return environment.get_value('FUZZ_DATA')

  local_data_bundle_directory = os.path.join(local_data_bundles_directory,
                                             data_bundle.name)

  if data_bundle.is_local:
    # Data bundle is on local disk, return path.
    return local_data_bundle_directory

  # This data bundle is on NFS, calculate path.
  # Make sure that NFS_ROOT pointing to nfs server is set. If not, use local.
  if not environment.get_value('NFS_ROOT'):
    logs.log_warn('NFS_ROOT is not set, using local corpora directory.')
    return local_data_bundle_directory

  return _get_nfs_data_bundle_path(data_bundle.name)


def get_fuzzer_directory(fuzzer_name):
  """Return directory used by a fuzzer."""
  builtin_fuzzer = builtin_fuzzers.get(fuzzer_name)
  if builtin_fuzzer:
    return builtin_fuzzer.fuzzer_directory

  fuzzer_directory = environment.get_value('FUZZERS_DIR')
  fuzzer_directory = os.path.join(fuzzer_directory, fuzzer_name)
  return fuzzer_directory


def is_directory_on_nfs(data_bundle_directory):
  """Return whether this directory is on NFS."""
  nfs_root = environment.get_value('NFS_ROOT')
  if not nfs_root:
    return False

  data_bundle_directory_real_path = os.path.realpath(data_bundle_directory)
  nfs_root_real_path = os.path.realpath(nfs_root)
  return data_bundle_directory_real_path.startswith(nfs_root_real_path + os.sep)


def archive_testcase_and_dependencies_in_gcs(resource_list, testcase_path):
  """Archive testcase and its dependencies, and store in blobstore."""
  if not os.path.exists(testcase_path):
    logs.log_error('Unable to find testcase %s.' % testcase_path)
    return None, None, None, None

  absolute_filename = testcase_path
  archived = False
  zip_filename = None
  zip_path = None

  if not resource_list:
    resource_list = []

  # Add resource dependencies based on testcase path. These include
  # stuff like extensions directory, dependency files, etc.
  resource_list.extend(
      testcase_manager.get_resource_dependencies(testcase_path))

  # Filter out duplicates, directories, and files that do not exist.
  resource_list = utils.filter_file_list(resource_list)

  logs.log('Testcase and related files :\n%s' % str(resource_list))

  if len(resource_list) <= 1:
    # If this does not have any resources, just save the testcase.
    # TODO(flowerhack): Update this when we teach CF how to download testcases.
    try:
      file_handle = open(testcase_path, 'rb')
    except IOError:
      logs.log_error('Unable to open testcase %s.' % testcase_path)
      return None, None, None, None
  else:
    # If there are resources, create an archive.

    # Find the common root directory for all of the resources.
    # Assumption: resource_list[0] is the testcase path.
    base_directory_list = resource_list[0].split(os.path.sep)
    for list_index in range(1, len(resource_list)):
      current_directory_list = resource_list[list_index].split(os.path.sep)
      length = min(len(base_directory_list), len(current_directory_list))
      for directory_index in range(length):
        if (current_directory_list[directory_index] !=
            base_directory_list[directory_index]):
          base_directory_list = base_directory_list[0:directory_index]
          break

    base_directory = os.path.sep.join(base_directory_list)
    logs.log('Subresource common base directory: %s' % base_directory)
    if base_directory:
      # Common parent directory, archive sub-paths only.
      base_len = len(base_directory) + len(os.path.sep)
    else:
      # No common parent directory, archive all paths as it-is.
      base_len = 0

    # Prepare the filename for the archive.
    zip_filename, _ = os.path.splitext(os.path.basename(testcase_path))
    zip_filename += _TESTCASE_ARCHIVE_EXTENSION

    # Create the archive.
    zip_path = os.path.join(environment.get_value('INPUT_DIR'), zip_filename)
    zip_file = zipfile.ZipFile(zip_path, 'w')
    for file_name in resource_list:
      if os.path.exists(file_name):
        relative_filename = file_name[base_len:]
        zip_file.write(file_name, relative_filename, zipfile.ZIP_DEFLATED)
    zip_file.close()

    try:
      file_handle = open(zip_path, 'rb')
    except IOError:
      logs.log_error('Unable to open testcase archive %s.' % zip_path)
      return None, None, None, None

    archived = True
    absolute_filename = testcase_path[base_len:]

  fuzzed_key = blobs.write_blob(file_handle)
  file_handle.close()

  # Don't need the archive after writing testcase to blobstore.
  if zip_path:
    shell.remove_file(zip_path)

  return fuzzed_key, archived, absolute_filename, zip_filename



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for finding similar issues."""

from flask import request

from handlers import base_handler
from libs import handler
from libs import helpers
from libs.issue_management import issue_tracker_utils


class Handler(base_handler.Handler):
  """Handler that finds similar issues."""

  @staticmethod
  def get_issues(issue_tracker, testcase, only_open):
    """Get similar issues. It is used by self.process() and
    handler.testcase_detail.FindSimilarIssuesHandler.get()"""
    issues = issue_tracker_utils.get_similar_issues(
        issue_tracker, testcase, only_open=only_open)

    items = []
    for entry in issues:
      items.append({
          'owner': entry.assignee,
          'reporter': entry.reporter,
          'status': entry.status,
          'title': entry.title,
          'id': entry.id
      })

    items = sorted(items, key=lambda k: k['id'])
    return [{
        'issue': item,
        'url': issue_tracker.issue_url(item['id']),
    } for item in items]

  @handler.get(handler.JSON)
  @handler.check_admin_access_if_oss_fuzz
  @handler.check_testcase_access
  def get(self, testcase):
    """Find similar issues."""
    filter_type = request.get('filterType')
    only_open = filter_type == 'open'

    issue_tracker = helpers.get_issue_tracker_for_testcase(testcase)
    items = self.get_issues(issue_tracker, testcase, only_open)

    response = {
        'queryString':
            ' '.join(issue_tracker_utils.get_search_keywords(testcase)),
        'queryUrl':
            issue_tracker_utils.get_similar_issues_url(issue_tracker, testcase,
                                                       only_open),
        'items':
            items,
    }
    return self.render_json(response)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common utility functions."""

import ast
import datetime
import functools
import gc
import hashlib
import inspect
import os
import random
import sys
import time
import urllib.parse
import urllib.request
import weakref

import requests

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import retry
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

try:
  import psutil
except ImportError:
  psutil = None

# FIXME: Binary extensions list is still very basic.
BINARY_EXTENSIONS = [
    # Media formats.
    '.mp3',
    '.ogg',
    '.mp4',
    '.webm',
    # Image Formats.
    '.png',
    '.jpg',
    '.gif',
    # Misc.
    '.pdf',
    '.swf',
]
FUZZ_PREFIX = 'fuzz-'
TEXT_EXTENSIONS = [
    '.css', '.js', '.htm', '.html', '.svg', '.xhtml', '.xht', '.xml', '.xsl'
]
URL_REQUEST_RETRIES = 5
URL_REQUEST_FAIL_WAIT = 1
WINDOWS_PREFIX_PATH = '\\\\?\\'

# Thread pool for use in function timeouts.
THREAD_POOL = None

LOCAL_SOURCE_MANIFEST = os.path.join('src', 'appengine', 'resources',
                                     'clusterfuzz-source.manifest')


def utcnow():
  """Return datetime.datetime.utcnow(). We need this method because we can't
    mock built-in methods."""
  return datetime.datetime.utcnow()  # pragma: no cover.


def current_date_time():
  """Returns current date and time."""
  return datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')


def utc_date_to_timestamp(date):
  """Converts a (UTC) datetime.date to a UNIX timestamp."""
  return (date - datetime.date(1970, 1, 1)).total_seconds()


def utc_datetime_to_timestamp(dt):
  """Converts a (UTC) datetime.date to a UNIX timestamp."""
  return (dt - datetime.datetime.utcfromtimestamp(0)).total_seconds()


def decode_to_unicode(obj):
  """Decode object to unicode encoding."""
  if not hasattr(obj, 'decode'):
    return obj

  return obj.decode('utf-8', errors='ignore')


def encode_as_unicode(obj):
  """Encode a string as unicode, or leave bytes as they are."""
  if not hasattr(obj, 'encode'):
    return obj

  return obj.encode('utf-8')


@retry.wrap(
    retries=URL_REQUEST_RETRIES,
    delay=URL_REQUEST_FAIL_WAIT,
    function='base.utils.fetch_url')
def fetch_url(url):
  """Fetch url content."""
  operations_timeout = environment.get_value('URL_BLOCKING_OPERATIONS_TIMEOUT')

  response = requests.get(url, timeout=operations_timeout)
  if response.status_code == 404:
    return None

  response.raise_for_status()
  return response.text


def fields_match(string_1,
                 string_2,
                 field_separator=':',
                 allow_empty_fields=True):
  """Match fields of two strings, separated by a |field_separator|. Empty fields
  can be ignored via |allow_empty_fields| flag."""
  if string_1 is None or string_2 is None:
    return False
  if string_1 == string_2:
    return True

  string_1_fields = string_1.split(field_separator)
  string_2_fields = string_2.split(field_separator)

  if not allow_empty_fields and len(string_1_fields) != len(string_2_fields):
    return False

  min_fields_length = min(len(string_1_fields), len(string_2_fields))
  for i in range(min_fields_length):
    if string_1_fields[i] != string_2_fields[i]:
      return False

  return True


def file_path_to_file_url(path):
  """Return a path as a file scheme url."""
  if not path:
    return ''

  path = path.lstrip(WINDOWS_PREFIX_PATH)
  return urllib.parse.urljoin('file:', urllib.request.pathname2url(path))


def filter_file_list(file_list):
  """Filters file list by removing duplicates, non-existent files
  and directories."""
  filtered_file_list = []
  for file_path in file_list:
    if not os.path.exists(file_path):
      continue

    if os.path.isdir(file_path):
      continue

    # Do a os specific case normalization before comparison.
    if (os.path.normcase(file_path) in list(
        map(os.path.normcase, filtered_file_list))):
      continue

    filtered_file_list.append(file_path)

  if len(filtered_file_list) != len(file_list):
    logs.log('Filtered file list (%s) from (%s).' % (str(filtered_file_list),
                                                     str(file_list)))

  return filtered_file_list


def find_binary_path(app_directory, binary_file_subpath):
  """Find the path to a binary given the app directory and the file name.

  This is necessary as cov files are created in the root app directory, and we
  need a way to find the corresponding binary to symbolize addresses."""
  binary_path = os.path.join(app_directory, binary_file_subpath)
  if os.path.exists(binary_path):
    # Common case: the binary exists in the root directory.
    return binary_path

  # Match the longest file sub-path suffix.
  binary_file_subpath_with_sep = binary_file_subpath
  if not binary_file_subpath_with_sep.startswith(os.sep):
    binary_file_subpath_with_sep = os.sep + binary_file_subpath_with_sep
  for root, _, filenames in os.walk(app_directory):
    for filename in filenames:
      file_path = os.path.join(root, filename)
      if file_path.endswith(binary_file_subpath_with_sep):
        return file_path

  # Otherwise, do a search for the filename.
  binary_filename = os.path.basename(binary_file_subpath)
  for root, _, filenames in os.walk(app_directory):
    for filename in filenames:
      if filename == binary_filename:
        file_path = os.path.join(root, filename)
        return file_path

  return None


def get_application_id():
  """Return application id. Code simplified based off original implementation in
  AppEngine SDK get_identity.get_application_id."""
  app_id = environment.get_value('APPLICATION_ID')
  if app_id is None:
    return None

  psep = app_id.find('~')
  if psep > 0:
    app_id = app_id[psep + 1:]

  return app_id


def service_account_email():
  """Get the service account name."""
  # TODO(ochang): Detect GCE and return the GCE service account instead.
  email_id = get_application_id()
  if ':' in email_id:
    domain, application_id = email_id.split(':')
    email_id = application_id + '.' + domain

  return email_id + '@appspot.gserviceaccount.com'


def get_bot_testcases_file_path(input_directory):
  """Returns path to bot-specific fuzzed testcases."""
  # Using |FUZZ_INPUTS| prevents putting high load on nfs servers for cases
  # when |input_directory| is a cloud storage data bundle. We can't rely
  # on |FUZZ_INPUTS| always since it might not be available during local fuzzer
  # testing, so use |input_directory| if it is not defined.
  local_testcases_directory = environment.get_value('FUZZ_INPUTS')
  bot_testcases_directory = (
      local_testcases_directory
      if local_testcases_directory else input_directory)

  bot_name = environment.get_value('BOT_NAME')
  bot_testcases_filename = '.%s_testcases' % bot_name
  bot_testcases_file_path = os.path.join(bot_testcases_directory,
                                         bot_testcases_filename)

  return bot_testcases_file_path


def get_crash_stacktrace_output(application_command_line,
                                symbolized_stacktrace,
                                unsymbolized_stacktrace=None,
                                build_type=None):
  """Return output string with symbolized and unsymbolized stacktraces
  combined."""

  def _guess_build_type(application_command_line):
    if 'stable' in application_command_line:
      return 'stable'
    if 'beta' in application_command_line:
      return 'beta'
    if sub_string_exists_in(['debug', 'dbg'], application_command_line):
      return 'debug'
    return 'release'

  separator = '-' * 40
  if not build_type:
    build_type = _guess_build_type(application_command_line)

  crash_stacktraces_output = environment.get_environment_settings_as_string()
  if application_command_line:
    crash_stacktraces_output += (
        '[Command line] %s\n\n' % application_command_line)
  crash_stacktraces_output += ('+%s%s Build Stacktrace%s+\n%s' % (
      separator, build_type.capitalize(), separator, symbolized_stacktrace))

  # No unsymbolized stack available. Bail out.
  if not unsymbolized_stacktrace:
    return crash_stacktraces_output

  unsymbolized_stacktrace_diff = get_unique_lines_in_unsymbolized_stack(
      symbolized_stacktrace, unsymbolized_stacktrace)
  if unsymbolized_stacktrace_diff:
    crash_stacktraces_output += (
        '\n\n+%s%s Build Unsymbolized Stacktrace (diff)%s+\n\n%s' %
        (separator, build_type.capitalize(), separator,
         unsymbolized_stacktrace_diff))
  return crash_stacktraces_output


def get_directory_hash_for_path(file_path):
  """Return the directory hash for a file path (excludes file name)."""
  root_directory = environment.get_value('ROOT_DIR')

  directory_path = os.path.dirname(file_path)
  normalized_directory_path = remove_prefix(directory_path,
                                            root_directory + os.sep)
  normalized_directory_path = normalized_directory_path.replace('\\', '/')
  return string_hash(normalized_directory_path)


def get_file_contents_with_fatal_error_on_failure(path):
  """Return the contents of the specified file, or None on error."""
  try:
    with open(path, 'rb') as file_handle:
      data = file_handle.read()
    return data
  except IOError:
    logs.log_error('Unable to read file `%s\'' % path)

  raise errors.BadStateError


def get_line_seperator(label=''):
  """Return a line separator with an optional label."""
  separator = '-' * 40
  result = '\n\n%s%s%s\n\n' % (separator, label, separator)
  return result


def get_normalized_relative_path(file_path, directory_path):
  """Return normalized relative path for file w.r.t to a directory."""
  normalized_relative_file_path = remove_prefix(file_path,
                                                directory_path + os.sep)
  normalized_relative_file_path = (
      normalized_relative_file_path.replace('\\', '/'))
  return normalized_relative_file_path


def get_path_without_ext(path):
  """Return a path excluding the extension."""
  return os.path.splitext(path)[0]


def get_process_ids(process_id, recursive=True):
  """Return list of pids for a process and its descendants."""
  # Try to find the running process.
  if not psutil.pid_exists(process_id):
    return []

  pids = [process_id]

  try:
    psutil_handle = psutil.Process(process_id)
    children = psutil_handle.children(recursive=recursive)
    for child in children:
      pids.append(child.pid)
  except psutil.NoSuchProcess:
    # Avoid too much logging when the process already died.
    return []

  except (psutil.AccessDenied, OSError):
    logs.log_warn('Failed to get process children.')
    return []

  return pids


def get_line_count_string(line_count):
  """Return string representation for size."""
  if line_count == 0:
    return 'empty'
  if line_count == 1:
    return '1 line'
  return '%d lines' % line_count


def get_size_string(size):
  """Return string representation for size."""
  if size < 1 << 10:
    return '%d B' % size
  if size < 1 << 20:
    return '%d KB' % (size >> 10)
  if size < 1 << 30:
    return '%d MB' % (size >> 20)
  return '%d GB' % (size >> 30)


def get_unique_lines_in_unsymbolized_stack(symbolized_stacktrace,
                                           unsymbolized_stacktrace):
  """Return unique lines in unsymbolized stacktrace that are not in the
  symbolized stacktrace."""
  if symbolized_stacktrace == unsymbolized_stacktrace:
    return ''

  symbolized_stacktrace_lines = symbolized_stacktrace.splitlines()
  unsymbolized_stacktrace_lines = unsymbolized_stacktrace.splitlines()
  stripped_symbolized_stacktrace_lines = set()
  for line in symbolized_stacktrace_lines:
    stripped_symbolized_stacktrace_lines.add(line.strip())

  index = 0
  last_index = len(unsymbolized_stacktrace_lines) - 1
  start = -1
  end = -1
  while index <= last_index:
    if (unsymbolized_stacktrace_lines[index].strip() not in
        stripped_symbolized_stacktrace_lines):
      if start == -1:
        start = index
        end = index + 1
      else:
        end = index

    index += 1

  if start == -1:
    # Nothing unique found, return empty string.
    return ''

  line_gap = 2
  start = max(0, start - line_gap)
  end = min(end + line_gap, last_index + 1)
  result = '\n'.join(unsymbolized_stacktrace_lines[start:end])
  return result


def indent_string(string, chars):
  """Indents a string by x number of characters."""

  indented_string = ''
  for line in string.splitlines():
    indented_string += '%s%s\n' % ((' ' * chars), line)

  # Strip the ending '\n' and return result.
  return indented_string[0:-1]


def is_binary_file(file_path, bytes_to_read=1024):
  """Return true if the file looks like a binary file."""
  file_extension = os.path.splitext(file_path)[1].lower()
  if file_extension in BINARY_EXTENSIONS:
    return True
  if file_extension in TEXT_EXTENSIONS:
    return False

  text_characters = list(map(chr, list(range(32, 128)))) + ['\r', '\n', '\t']
  try:
    with open(file_path, 'rb') as file_handle:
      data = file_handle.read(bytes_to_read)
  except:
    logs.log_error('Could not read file %s in is_binary_file.' % file_path)
    return None

  binary_data = [char for char in data if char not in text_characters]

  return len(binary_data) > len(data) * 0.1


def is_recursive_call():
  """Returns true if the caller function is called recursively."""
  try:
    stack_frames = inspect.stack()
    caller_name = stack_frames[1][3]
    for stack_frame_index in range(2, len(stack_frames)):
      if caller_name == stack_frames[stack_frame_index][3]:
        return True
  except:
    pass

  return False


def is_valid_testcase_file(file_path,
                           check_if_exists=True,
                           size_limit=None,
                           allowed_extensions=None):
  """Return true if the file looks like a testcase file."""
  filename = os.path.basename(file_path)
  if filename.startswith('.') or filename.startswith(FUZZ_PREFIX):
    return False

  if allowed_extensions:
    file_extension = os.path.splitext(file_path)[1].lower()
    if file_extension not in allowed_extensions:
      return False

  directories_to_ignore = ['.git', '.hg', '.svn']
  for directory_to_ignore in directories_to_ignore:
    directory_string = '%s%s%s' % (os.sep, directory_to_ignore, os.sep)
    if directory_string in file_path:
      return False

  if (check_if_exists or size_limit) and not os.path.exists(file_path):
    return False

  if size_limit and os.path.getsize(file_path) > size_limit:
    return False

  return True


def maximum_parallel_processes_allowed():
  """Return maximum number of parallel processes allowed. Adjust it based
  on thread multiplier."""
  if environment.is_trusted_host():
    # gRPC only supports 1 thread/process.
    return 1

  max_parallel_process_count = environment.get_value('MAX_FUZZ_THREADS', 1)
  thread_multiplier = environment.get_value('THREAD_MULTIPLIER', 1)

  max_parallel_process_count *= thread_multiplier
  return int(max_parallel_process_count)


def normalize_path(path):
  """Normalize path. This is needed on windows because windows' paths are
    case-insensitive."""
  return os.path.normcase(os.path.normpath(path))


def python_gc():
  """Call python's garbage collector."""
  # gc_collect isn't perfectly synchronous, because it may
  # break reference cycles that then take time to fully
  # finalize. Call it thrice and hope for the best.
  for _ in range(3):
    gc.collect()


def random_element_from_list(element_list):
  """Returns a random element from list."""
  return element_list[random.SystemRandom().randint(0, len(element_list) - 1)]


def random_number(start, end):
  """Returns a random number between start and end."""
  return random.SystemRandom().randint(start, end)


# pylint: disable=inconsistent-return-statements
def random_weighted_choice(element_list, weight_attribute='weight'):
  """Returns a random element from list taking its weight into account."""
  total = sum(getattr(e, weight_attribute) for e in element_list)
  random_pick = random.SystemRandom().uniform(0, total)
  temp = 0
  for element in element_list:
    element_weight = getattr(element, weight_attribute)
    if element_weight == 0:
      continue
    if temp + element_weight >= random_pick:
      return element
    temp += element_weight

  assert False, 'Failed to make a random weighted choice.'


def read_data_from_file(file_path, eval_data=True, default=None):
  """Returns eval-ed data from file."""
  if not os.path.exists(file_path):
    return default

  failure_wait_interval = environment.get_value('FAIL_WAIT')
  file_content = None
  retry_limit = environment.get_value('FAIL_RETRIES')
  for _ in range(retry_limit):
    try:
      with open(file_path, 'rb') as file_handle:
        file_content = file_handle.read()
    except:
      file_content = None
      logs.log_warn('Error occurred while reading %s, retrying.' % file_path)
      time.sleep(random.uniform(1, failure_wait_interval))
      continue

  if file_content is None:
    logs.log_error('Failed to read data from file %s.' % file_path)
    return None

  if not eval_data:
    return file_content

  if not file_content:
    return default

  try:
    return ast.literal_eval(file_content.decode('utf-8'))
  except (SyntaxError, TypeError):
    return None


def remove_prefix(string, prefix):
  """Strips the prefix from a string."""
  if string.startswith(prefix):
    return string[len(prefix):]

  return string


def remove_sub_strings(string, substrings):
  """Strips substrings from a given string."""
  result = string
  for substring in substrings:
    result = result.replace(substring, '')

  return result


def restart_machine():
  """Restart machine."""
  if environment.platform() == 'WINDOWS':
    os.system('shutdown /f /r /t 0')
  else:
    # POSIX platforms.
    os.system('sudo shutdown -r now')


def search_bytes_in_file(search_bytes, file_handle):
  """Helper to search for bytes in a large binary file without memory
  issues.
  """
  # TODO(aarya): This is too brittle and will fail if we have a very large
  # line.
  for line in file_handle:
    if search_bytes in line:
      return True

  return False


def string_hash(obj):
  """Returns a SHA-1 hash of the object. Not used for security purposes."""
  return hashlib.sha1(str(obj).encode('utf-8')).hexdigest()


def entity_hash(obj):
  """Returns a deterministic hash of a ndb entity.

  If an entity has been recently modified, put() must be called on it before
  this function will pick up the changes.
  """
  hasher = hashlib.sha1()
  entity_dict = obj.to_dict()
  for key in sorted(entity_dict.keys()):
    hasher.update(str(entity_dict[key]).encode('utf-8'))

  return hasher.hexdigest()


def string_is_true(value):
  """Check to see if a string has a value that should be treated as True."""
  return value and value != 'false' and value != 'False' and value != '0'


def strip_from_left(string, prefix):
  """Strip a prefix from start from string."""
  if not string.startswith(prefix):
    return string
  return string[len(prefix):]


def strip_from_right(string, suffix):
  """Strip a suffix from end of string."""
  if not string.endswith(suffix):
    return string
  return string[:len(string) - len(suffix)]


def sub_string_exists_in(substring_list, string):
  """Return true if one of the substring in the list is found in |string|."""
  for substring in substring_list:
    if substring in string:
      return True

  return False


def time_difference_string(timestamp):
  """Return time difference as a string."""
  if not timestamp:
    return ''

  delta = int((datetime.datetime.utcnow() - timestamp).total_seconds())
  d_minutes = delta // 60
  d_hours = d_minutes // 60
  d_days = d_hours // 24

  if d_days > 6:
    return '%s' % str(timestamp).split()[0]
  if d_days > 1:
    return '%s days ago' % d_days  # starts at 2 days.
  if d_hours > 1:
    return '%s hours ago' % d_hours  # starts at 2 hours.
  if d_minutes > 1:
    return '%s minutes ago' % d_minutes
  if d_minutes > 0:
    return '1 minute ago'
  if delta > -30:
    return 'moments ago'

  # Only say something is in the future if it is more than just clock skew.
  return 'in the future'


def timeout(duration):
  """Timeout decorator for functions."""

  def decorator(func):
    """Decorates the given function."""
    if environment.is_running_on_app_engine():
      # multiprocessing doesn't work on App Engine.
      return func

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
      """Wrapper."""
      # FIXME: Weird exceptions in imports, might be something relating to our
      # reload module. Needs further investigation, try this as a temporary fix.
      import multiprocessing.pool
      import threading

      # Fix for Python < 2.7.2.
      if not hasattr(threading.current_thread(), '_children'):
        # pylint: disable=protected-access
        threading.current_thread()._children = weakref.WeakKeyDictionary()

      global THREAD_POOL
      if THREAD_POOL is None:
        THREAD_POOL = multiprocessing.pool.ThreadPool(processes=3)

      try:
        from clusterfuzz._internal.datastore import \
            ndb_init  # Avoid circular import.
        async_result = THREAD_POOL.apply_async(
            ndb_init.thread_wrapper(func), args=args, kwds=kwargs)
        return async_result.get(timeout=duration)
      except multiprocessing.TimeoutError:
        # Sleep for some minutes in order to wait for flushing metrics.
        time.sleep(120)

        # If we don't exit here, we will cause threads to pile up and leading to
        # out-of-memory. Safe to just exit here.
        logs.log_fatal_and_exit(
            ('Exception occurred in function {0}: args: {1}, kwargs: {2}'
             ' exception: {3}').format(func, args, kwargs,
                                       sys.exc_info()[1]))

    return _wrapper

  return decorator


def wait_until_timeout(threads, thread_timeout):
  """Wait for all threads to finish unless the given timeout is reached.

  If no thread is alive, it waits much shorter than the given timeout.

  Return True if timeout is exceeded, and return False otherwise.
  """
  thread_alive_check_interval = environment.get_value(
      'THREAD_ALIVE_CHECK_INTERVAL')
  if not thread_alive_check_interval:
    time.sleep(thread_timeout)
    return False

  wait_timeout = time.time() + thread_timeout
  while time.time() < wait_timeout:
    time.sleep(thread_alive_check_interval)

    thread_alive = False
    for thread in threads:
      if thread.is_alive():
        thread_alive = True
        break

    if not thread_alive:
      return False

  return True


def write_data_to_file(content, file_path, append=False):
  """Writes data to file."""
  failure_wait_interval = environment.get_value('FAIL_WAIT')
  file_mode = 'ab' if append else 'wb'
  retry_limit = environment.get_value('FAIL_RETRIES')

  # TODO(mbarbella): One extra iteration is allowed for the type conversion hack
  # included here. Once this function is converted to only accept bytes-like
  # objects, it should be adjusted back to the normal retry limit.
  for _ in range(retry_limit + 1):
    try:
      with open(file_path, file_mode) as file_handle:
        file_handle.write(content)
    except TypeError:
      # If we saw a TypeError, content was not bytes-like. Convert it.
      content = str(content).encode('utf-8')
      continue
    except EnvironmentError:
      # An EnvironmentError signals a problem writing the file. Retry in case
      # it was a spurious error.
      logs.log_warn('Error occurred while writing %s, retrying.' % file_path)
      time.sleep(random.uniform(1, failure_wait_interval))
      continue

    # Successfully written data file.
    return

  logs.log_error('Failed to write data to file %s.' % file_path)


@memoize.wrap(memoize.FifoInMemory(1))
def default_backup_bucket():
  """Return the default backup bucket for this instance of ClusterFuzz."""
  # Do not use |BACKUP_BUCKET| environment variable as that is the overridden
  # backup bucket from job type and is not the default backup bucket.
  return local_config.ProjectConfig().get('env.BACKUP_BUCKET')


@memoize.wrap(memoize.FifoInMemory(1))
def default_project_name():
  """Return the default project name for this instance of ClusterFuzz."""
  # Do not use |PROJECT_NAME| environment variable as that is the overridden
  # project name from job type and is not the default project name.
  return local_config.ProjectConfig().get('env.PROJECT_NAME')


def current_project():
  """Return the project for the current job, or the default project."""
  return environment.get_value('PROJECT_NAME', default_project_name())


def current_source_version():
  """Return the current source revision."""
  # For test use.
  source_version_override = environment.get_value('SOURCE_VERSION_OVERRIDE')
  if source_version_override:
    return source_version_override

  root_directory = environment.get_value('ROOT_DIR')
  local_manifest_path = os.path.join(root_directory, LOCAL_SOURCE_MANIFEST)
  if os.path.exists(local_manifest_path):
    return read_data_from_file(
        local_manifest_path, eval_data=False).strip().decode('utf-8')

  return None


def read_from_handle_truncated(file_handle, max_len):
  """Read from file handle, limiting output to |max_len| by removing output in
  the middle."""
  file_handle.seek(0, os.SEEK_END)
  file_size = file_handle.tell()
  file_handle.seek(0, os.SEEK_SET)

  if file_size <= max_len:
    return file_handle.read()

  # Read first and last |half_max_len| bytes.
  half_max_len = max_len // 2
  start = file_handle.read(half_max_len)
  file_handle.seek(file_size - half_max_len, os.SEEK_SET)
  end = file_handle.read(half_max_len)

  truncated_marker = b'\n...truncated %d bytes...\n' % (file_size - max_len)

  return start + truncated_marker + end


def normalize_email(email):
  """Normalize an email address."""
  # TODO(ochang): Investigate whether if it makes sense to replace
  # @googlemail.com with @gmail.com.
  return email.lower()


def emails_equal(first, second):
  """Return whether or not the 2 emails are equal after being normalized."""
  if not first or not second:
    return False

  return normalize_email(first) == normalize_email(second)


def parse_delimited(value_or_handle, delimiter, strip=False,
                    remove_empty=False):
  """Parse a delimter separated value."""
  if hasattr(value_or_handle, 'read'):
    results = value_or_handle.read().split(delimiter)
  else:
    results = value_or_handle.split(delimiter)

  if not strip and not remove_empty:
    return results

  processed_results = []
  for result in results:
    if strip:
      result = result.strip()

    if remove_empty and not result:
      continue

    processed_results.append(result)

  return processed_results


def is_oss_fuzz():
  """If this is an instance of OSS-Fuzz."""
  return default_project_name() == 'oss-fuzz'


def is_chromium():
  """If this is an instance of chromium fuzzing."""
  return default_project_name() == 'chromium'


def file_hash(file_path):
  """Returns the SHA-1 hash of |file_path| contents."""
  chunk_size = 51200  # Read in 50 KB chunks.
  digest = hashlib.sha1()
  with open(file_path, 'rb') as file_handle:
    chunk = file_handle.read(chunk_size)
    while chunk:
      digest.update(chunk)
      chunk = file_handle.read(chunk_size)

  return digest.hexdigest()


def cpu_count():
  """Get the CPU count."""
  # Does not import on App Engine.
  import multiprocessing

  return environment.get_value('CPU_COUNT_OVERRIDE',
                               multiprocessing.cpu_count())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""request_cache tests."""

import unittest

from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from libs import request_cache


class CacheClass(object):
  """Test cache class."""

  def __init__(self):
    self.called = []

  @request_cache.wrap(3)
  def foo(self, a):
    self.called.append((a,))
    return a + 1


class CacheClass2(object):
  """Test cache class."""

  def __init__(self):
    self.called = []

  @request_cache.wrap(3)
  def foo(self, a):
    self.called.append((a,))
    return a + 2


class FakeRequest(object):
  """Fake request."""


class CacheTest(unittest.TestCase):
  """Cache tests."""

  def setUp(self):
    test_helpers.patch(self, ['libs.request_cache.get_cache_backing'])

    self.mock.get_cache_backing.return_value = FakeRequest()

  def test_basic(self):
    """Basic tests."""
    c = CacheClass()
    self.assertEqual(2, c.foo(1))
    self.assertEqual(2, c.foo(1))
    self.assertEqual(3, c.foo(2))

    self.assertListEqual([
        (1,),
        (2,),
    ], c.called)

  def test_no_request(self):
    """Test no request available."""
    self.mock.get_cache_backing.return_value = None
    c = CacheClass()
    self.assertEqual(2, c.foo(1))
    self.assertEqual(2, c.foo(1))
    self.assertEqual(3, c.foo(2))

    self.assertListEqual([
        (1,),
        (1,),
        (2,),
    ], c.called)

  def test_name_clash(self):
    """Test name clash."""
    c1 = CacheClass()
    c2 = CacheClass2()
    self.assertEqual(2, c1.foo(1))
    self.assertEqual(3, c2.foo(1))

    self.assertListEqual([
        (1,),
    ], c1.called)

    self.assertListEqual([
        (1,),
    ], c2.called)

  def test_scoped_to_request(self):
    """Test that the cache is scoped to requests."""
    c = CacheClass()
    self.assertEqual(2, c.foo(1))
    self.mock.get_cache_backing.return_value = FakeRequest()
    self.assertEqual(2, c.foo(1))

    self.assertListEqual([
        (1,),
        (1,),
    ], c.called)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Issue tracker comments."""

import re


class Comment(object):
  """Class representing a single comment update."""

  def __init__(self):
    self.author = None
    self.cc = None
    self.comment = None
    self.components = []
    self.created = None
    self.labels = []
    self.summary = None
    self.status = None
    self.owner = None
    self.id = 0

  def has_label_containing(self, expression):
    return any(self.get_labels_containing(expression))

  def get_labels_containing(self, expression):
    regex = re.compile(expression, re.DOTALL | re.IGNORECASE)
    return [label for label in self.labels if regex.search(label)]

  def has_label_matching(self, expression):
    return any(self.get_labels_matching(expression))

  def get_labels_matching(self, expression):
    regex = re.compile(expression + r'\Z', re.DOTALL | re.IGNORECASE)
    return [label for label in self.labels if regex.match(label)]

  def get_labels_by_prefix(self, prefix):
    return self.get_labels_matching('%s.*' % prefix)

  def has_label(self, value):
    for label in self.labels:
      if label.lower() == value.lower():
        return True

    return False



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""reproduce.py reproduces test cases locally."""

from clusterfuzz._internal.base import modules

modules.fix_module_search_paths()

import os
import shutil
import tempfile
import time
from urllib import parse

from clusterfuzz._internal.base import json_utils
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers import init
from clusterfuzz._internal.bot.tasks import commands
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import archive
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell
from local.butler import appengine
from local.butler.reproduce_tool import android
from local.butler.reproduce_tool import config
from local.butler.reproduce_tool import errors
from local.butler.reproduce_tool import http_utils
from local.butler.reproduce_tool import prompts

CONFIG_DIRECTORY = os.path.join(
    os.path.expanduser('~'), '.config', 'clusterfuzz')
DISPLAY = ':99'
PROCESS_START_WAIT_SECONDS = 2
SUPPORTED_PLATFORMS = ['android', 'fuchsia', 'linux', 'mac']

FILENAME_RESPONSE_HEADER = 'x-goog-meta-filename'


class SerializedTestcase(object):
  """Minimal representation of a test case."""

  def __init__(self, testcase_map):
    self._testcase_map = testcase_map

  def __getattr__(self, item):
    return self._testcase_map[item]

  def get_metadata(self, key=None, default=None):
    """Emulate Testcase's get_metadata function."""
    metadata = json_utils.loads(self.additional_metadata)
    if not key:
      return metadata

    try:
      return self.metadata[key]
    except KeyError:
      return default

  def actual_fuzzer_name(self):
    """Actual fuzzer name, uses one from overridden attribute if available."""
    return self.overridden_fuzzer_name or self.fuzzer_name

  def get_fuzz_target(self):
    """Return a fuzz target for this test case in the expected format."""
    if not self.serialized_fuzz_target:
      return None

    fuzz_target = data_types.FuzzTarget(
        engine=self.serialized_fuzz_target['engine'],
        project=self.serialized_fuzz_target['project'],
        binary=self.serialized_fuzz_target['binary'])
    return fuzz_target


def _get_testcase(testcase_id, configuration):
  """Retrieve the json representation of the test case with the given id."""
  response, content = http_utils.request(
      configuration.get('testcase_info_url'),
      body={'testcaseId': testcase_id},
      configuration=configuration)

  if response.status != 200:
    raise errors.ReproduceToolUnrecoverableError(
        'Unable to fetch test case information.')

  testcase_map = json_utils.loads(content)
  return SerializedTestcase(testcase_map)


def _download_testcase(testcase_id, testcase, configuration):
  """Download the test case and return its path."""
  print('Downloading testcase...')
  testcase_download_url = '{url}?id={id}'.format(
      url=configuration.get('testcase_download_url'), id=testcase_id)
  response, content = http_utils.request(
      testcase_download_url,
      method=http_utils.GET_METHOD,
      configuration=configuration)

  if response.status != 200:
    raise errors.ReproduceToolUnrecoverableError(
        'Unable to download test case.')

  bot_absolute_filename = response[FILENAME_RESPONSE_HEADER]
  # Store the test case in the config directory for debuggability.
  testcase_directory = os.path.join(CONFIG_DIRECTORY, 'current-testcase')
  shell.remove_directory(testcase_directory, recreate=True)
  environment.set_value('FUZZ_INPUTS', testcase_directory)
  testcase_path = os.path.join(testcase_directory,
                               os.path.basename(bot_absolute_filename))

  utils.write_data_to_file(content, testcase_path)

  # Unpack the test case if it's archived.
  # TODO(mbarbella): Rewrite setup.unpack_testcase and share this code.
  if testcase.minimized_keys and testcase.minimized_keys != 'NA':
    mask = data_types.ArchiveStatus.MINIMIZED
  else:
    mask = data_types.ArchiveStatus.FUZZED

  if testcase.archive_state & mask:
    archive.unpack(testcase_path, testcase_directory)
    file_list = archive.get_file_list(testcase_path)

    testcase_path = None
    for file_name in file_list:
      if os.path.basename(file_name) == os.path.basename(
          testcase.absolute_path):
        testcase_path = os.path.join(testcase_directory, file_name)
        break

    if not testcase_path:
      raise errors.ReproduceToolUnrecoverableError(
          'Test case file was not found in archive.\n'
          'Original filename: {absolute_path}.\n'
          'Archive contents: {file_list}'.format(
              absolute_path=testcase.absolute_path, file_list=file_list))

  return testcase_path


def _setup_x():
  """Start Xvfb and blackbox before running the test application."""
  if environment.platform() != 'LINUX':
    return []

  if environment.is_engine_fuzzer_job():
    # For engine fuzzer jobs like AFL, libFuzzer, Xvfb is not needed as the
    # those fuzz targets do not needed a UI.
    return []

  environment.set_value('DISPLAY', DISPLAY)

  print('Creating virtual display...')
  xvfb_runner = new_process.ProcessRunner('/usr/bin/Xvfb')
  xvfb_process = xvfb_runner.run(additional_args=[
      DISPLAY, '-screen', '0', '1280x1024x24', '-ac', '-nolisten', 'tcp'
  ])
  time.sleep(PROCESS_START_WAIT_SECONDS)

  blackbox_runner = new_process.ProcessRunner('/usr/bin/blackbox')
  blackbox_process = blackbox_runner.run()
  time.sleep(PROCESS_START_WAIT_SECONDS)

  # Return all handles we create so they can be terminated properly at exit.
  return [xvfb_process, blackbox_process]


def _prepare_initial_environment(build_directory, iterations, verbose):
  """Prepare common environment variables that don't depend on the job."""
  # Create a temporary directory to use as ROOT_DIR with a copy of the default
  # bot and configuration directories nested under it.
  root_dir = environment.get_value('ROOT_DIR')
  temp_root_dir = tempfile.mkdtemp()
  environment.set_value('ROOT_DIR', temp_root_dir)

  def _update_directory(directory_name, ignore_paths=None):
    """Copy a subdirectory from a checkout to a temp directory."""
    if not ignore_paths:
      ignore_paths = []

    shutil.copytree(
        os.path.join(root_dir, directory_name),
        os.path.join(temp_root_dir, directory_name),
        ignore=lambda directory, contents:
        contents if directory in ignore_paths else [])

  _update_directory('bot')
  _update_directory('configs')
  _update_directory('resources')
  _update_directory(
      'src',
      ignore_paths=[
          os.path.join(root_dir, 'src', 'appengine'),
          os.path.join(root_dir, 'src', 'bazel-bin'),
          os.path.join(root_dir, 'src', 'bazel-genfiles'),
          os.path.join(root_dir, 'src', 'bazel-out'),
          os.path.join(root_dir, 'src', 'bazel-src'),
          os.path.join(root_dir, 'src', 'clusterfuzz', '_internal', 'tests'),
      ])

  environment.set_value('CONFIG_DIR_OVERRIDE',
                        os.path.join(temp_root_dir, 'configs', 'test'))
  environment.set_value(
      'PYTHONPATH',
      os.pathsep.join(
          [os.path.join(temp_root_dir, 'src'),
           appengine.find_sdk_path()]))

  environment.set_bot_environment()

  # Overrides that should not be set to the default values.
  environment.set_value('APP_DIR', build_directory)
  environment.set_value('BUILD_DIR', build_directory)
  environment.set_value('BUILDS_DIR', build_directory)

  # Some functionality must be disabled when running the tool.
  environment.set_value('REPRODUCE_TOOL', True)

  environment.set_value('TASK_NAME', 'reproduce')

  # Force logging to console for this process and child processes.
  if verbose:
    environment.set_value('LOG_TO_CONSOLE', True)

  if iterations:
    environment.set_value('CRASH_RETRIES', iterations)


def _verify_target_exists(build_directory):
  """Ensure that we can find the test target before running it.

  Separated into its own function to simplify test behavior."""
  if not build_manager.check_app_path():
    raise errors.ReproduceToolUnrecoverableError(
        'Unable to locate app binary in {build_directory}.'.format(
            build_directory=build_directory))


def _update_environment_for_testcase(testcase, build_directory,
                                     application_override):
  """Update environment variables that depend on the test case."""
  commands.update_environment_for_job(testcase.job_definition)
  environment.set_value('JOB_NAME', testcase.job_type)

  # Override app name if explicitly specified.
  if application_override:
    environment.set_value('APP_NAME', application_override)

  if testcase.fuzzer_name:
    fuzzer_directory = setup.get_fuzzer_directory(testcase.fuzzer_name)
  else:
    fuzzer_directory = os.path.join(environment.get_value('ROOT_DIR'), 'fuzzer')
    shell.create_directory(fuzzer_directory)

  environment.set_value('FUZZER_DIR', fuzzer_directory)

  task_name = environment.get_value('TASK_NAME')
  setup.prepare_environment_for_testcase(testcase, testcase.job_type, task_name)

  build_manager.set_environment_vars(
      [environment.get_value('FUZZER_DIR'), build_directory])

  _verify_target_exists(build_directory)


def _get_testcase_id_from_url(testcase_url):
  """Convert a testcase URL to a testcase ID."""
  url_parts = parse.urlparse(testcase_url)
  # Testcase urls have paths like "/testcase-detail/1234567890", where the
  # number is the testcase ID.
  path_parts = url_parts.path.split('/')

  try:
    testcase_id = int(path_parts[-1])
  except (ValueError, IndexError):
    testcase_id = 0

  # Validate that the URL is correct.
  if (len(path_parts) != 3 or path_parts[0] or
      path_parts[1] != 'testcase-detail' or not testcase_id):
    raise errors.ReproduceToolUnrecoverableError(
        'Invalid testcase URL {url}. Expected format: '
        'https://clusterfuzz-domain/testcase-detail/1234567890'.format(
            url=testcase_url))

  return testcase_id


def _print_stacktrace(result):
  """Display the output from a test case run."""
  print('#' * 80)
  print(result.get_stacktrace())
  print('#' * 80)
  print()


def _reproduce_crash(testcase_url, build_directory, iterations, disable_xvfb,
                     verbose, disable_android_setup, application):
  """Reproduce a crash."""
  _prepare_initial_environment(build_directory, iterations, verbose)

  # Validate the test case URL and fetch the tool's configuration.
  testcase_id = _get_testcase_id_from_url(testcase_url)
  configuration = config.ReproduceToolConfiguration(testcase_url)
  testcase = _get_testcase(testcase_id, configuration)

  # For new user uploads, we'll fail without the metadata set by analyze task.
  if not testcase.platform:
    raise errors.ReproduceToolUnrecoverableError(
        'This test case has not yet been processed. Please try again later.')

  # Ensure that we support this test case's platform.
  if testcase.platform not in SUPPORTED_PLATFORMS:
    raise errors.ReproduceToolUnrecoverableError(
        'The reproduce tool is not yet supported on {platform}.'.format(
            platform=testcase.platform))

  # Print warnings for this test case.
  if testcase.one_time_crasher_flag:
    print('Warning: this test case was a one-time crash. It may not be '
          'reproducible.')
  if testcase.flaky_stack:
    print('Warning: this test case is known to crash with different stack '
          'traces.')

  testcase_path = _download_testcase(testcase_id, testcase, configuration)
  _update_environment_for_testcase(testcase, build_directory, application)

  # Validate that we're running on the right platform for this test case.
  platform = environment.platform().lower()
  if testcase.platform == 'android' and platform == 'linux':
    android.prepare_environment(disable_android_setup)
  elif testcase.platform == 'android' and platform != 'linux':
    raise errors.ReproduceToolUnrecoverableError(
        'The ClusterFuzz environment only supports running Android test cases '
        'on Linux host machines. Unable to reproduce the test case on '
        '{current_platform}.'.format(current_platform=platform))
  elif testcase.platform != platform:
    raise errors.ReproduceToolUnrecoverableError(
        'The specified test case was discovered on {testcase_platform}. '
        'Unable to attempt to reproduce it on {current_platform}.'.format(
            testcase_platform=testcase.platform, current_platform=platform))

  x_processes = []
  if not disable_xvfb:
    _setup_x()
  timeout = environment.get_value('TEST_TIMEOUT')

  print('Running testcase...')
  try:
    result = testcase_manager.test_for_crash_with_retries(
        testcase, testcase_path, timeout, crash_retries=1)

    # If we can't reproduce the crash, prompt the user to try again.
    if not result.is_crash():
      _print_stacktrace(result)
      result = None
      use_default_retries = prompts.get_boolean(
          'Failed to find the desired crash on first run. Re-run '
          '{crash_retries} times?'.format(
              crash_retries=environment.get_value('CRASH_RETRIES')))
      if use_default_retries:
        print('Attempting to reproduce test case. This may take a while...')
        result = testcase_manager.test_for_crash_with_retries(
            testcase, testcase_path, timeout)

  except KeyboardInterrupt:
    print('Aborting...')
    result = None

  # Terminate Xvfb and blackbox.
  for process in x_processes:
    process.terminate()

  return result


def _cleanup():
  """Clean up after running the tool."""
  temp_directory = environment.get_value('ROOT_DIR')
  assert 'tmp' in temp_directory
  shell.remove_directory(temp_directory)


def execute(args):
  """Attempt to reproduce a crash then report on the result."""
  # Initialize fuzzing engines.
  init.run()

  # Prepare the emulator if needed.
  emulator_process = None
  if args.emulator:
    print('Starting emulator...')
    emulator_process = android.start_emulator()

  # The current working directory may change while we're running.
  absolute_build_dir = os.path.abspath(args.build_dir)
  try:
    result = _reproduce_crash(args.testcase, absolute_build_dir,
                              args.iterations, args.disable_xvfb, args.verbose,
                              args.disable_android_setup, args.application)
  except errors.ReproduceToolUnrecoverableError as exception:
    print(exception)
    return
  finally:
    if emulator_process:
      emulator_process.terminate()

  if not result:
    return

  _print_stacktrace(result)

  if result.is_crash():
    status_message = 'Test case reproduced successfully.'
  else:
    status_message = 'Unable to reproduce the desired crash.'
  print(status_message)

  _cleanup()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fuzzer.py."""

import os

from clusterfuzz._internal.bot.fuzzers.afl import fuzzer
from clusterfuzz._internal.tests.core.bot.fuzzers import builtin_test
from clusterfuzz._internal.tests.test_libs import helpers


class FuzzerTest(builtin_test.BaseEngineFuzzerTest):
  """Unit tests for fuzzer."""

  def setUp(self):
    super(FuzzerTest, self).setUp()
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.log_warn',
    ])

  def _test_passed(self):
    self.assertTrue(os.path.exists('/output/fuzz-0'))
    self.assertTrue(os.path.exists('/output/flags-0'))
    self.assertTrue(os.path.exists('/input/proj_target/in1'))

    with open('/output/flags-0') as f:
      self.assertEqual('%TESTCASE% target', f.read())

  def _test_failed(self):
    self.assertFalse(os.path.exists('/output/fuzz-0'))
    self.assertFalse(os.path.exists('/output/flags-0'))
    self.assertFalse(os.path.exists('/input/proj_target/in1'))

  def test_run(self):
    """Test running afl fuzzer."""
    afl = fuzzer.Afl()
    afl.run('/input', '/output', 1)

    self._test_passed()
